From 4420d39e512979855e765c8f0c14fa30ad861d88 Mon Sep 17 00:00:00 2001
From: AnusuyaKumarasamy <anusuya.kumarasamy@multicorewareinc.com>
Date: Fri, 19 Jul 2024 10:49:35 +0530
Subject: [PATCH 4/9] Added functions for IntaBlockCopy method in SCC

---
 source/common/cudata.cpp    |  308 +++++-
 source/common/cudata.h      |    9 +
 source/common/slice.cpp     |   53 ++
 source/common/slice.h       |    5 +
 source/encoder/analysis.cpp |  472 ++++++++-
 source/encoder/analysis.h   |   11 +
 source/encoder/dpb.cpp      |   42 +
 source/encoder/entropy.cpp  |    2 +-
 source/encoder/search.cpp   | 1796 ++++++++++++++++++++++++++++++++++-
 source/encoder/search.h     |   22 +
 source/x265.h               |    4 +
 11 files changed, 2706 insertions(+), 18 deletions(-)

diff --git a/source/common/cudata.cpp b/source/common/cudata.cpp
index ddee2b52b..a671df113 100644
--- a/source/common/cudata.cpp
+++ b/source/common/cudata.cpp
@@ -290,6 +290,8 @@ void CUData::initCTU(const Frame& frame, uint32_t cuAddr, int qp, uint32_t first
     m_bFirstRowInSlice = (uint8_t)firstRowInSlice;
     m_bLastRowInSlice  = (uint8_t)lastRowInSlice;
     m_bLastCuInSlice   = (uint8_t)lastCuInSlice;
+    m_lastIntraBCMv[0].set(0, 0);
+    m_lastIntraBCMv[1].set(0, 0);
 
     /* sequential memsets */
     m_partSet((uint8_t*)m_qp, (uint8_t)qp);
@@ -415,6 +417,8 @@ void CUData::copyPartFrom(const CUData& subCU, const CUGeom& childGeom, uint32_t
         memcpy(m_trCoeff[1] + tmpC2, subCU.m_trCoeff[1], sizeof(coeff_t) * tmpC);
         memcpy(m_trCoeff[2] + tmpC2, subCU.m_trCoeff[2], sizeof(coeff_t) * tmpC);
     }
+    for (int i = 0; i < 2; i++)
+        m_lastIntraBCMv[i] = subCU.m_lastIntraBCMv[i];
 }
 
 /* If a sub-CU part is not present (off the edge of the picture) its depth and
@@ -1723,7 +1727,7 @@ int CUData::getPMV(InterNeighbourMV* neighbours, uint32_t picList, uint32_t refI
     bool validIndirect[MD_ABOVE_LEFT + 1];
 
 #if ENABLE_MULTIVIEW
-    if (m_slice->m_param->numViews > 1)
+    if (m_slice->m_param->numViews > 1 || m_slice->m_param->bEnableSCC)
     {
         // Left candidate.
         if ((neighbours + MD_BELOW_LEFT)->isAvailable || (neighbours + MD_LEFT)->isAvailable)
@@ -1890,7 +1894,7 @@ int CUData::getPMV(InterNeighbourMV* neighbours, uint32_t picList, uint32_t refI
     }
 
     while (num < AMVP_NUM_CANDS)
-        amvpCand[num++] = 0;
+        amvpCand[num++].set(0, 0);
 
     return numMvc;
 }
@@ -2257,3 +2261,303 @@ void CUData::calcCTUGeoms(uint32_t ctuWidth, uint32_t ctuHeight, uint32_t maxCUS
         rangeCUIdx += sbWidth * sbWidth;
     }
 }
+
+
+bool CUData::getDerivedBV(uint32_t absPartIdx, const MV& currentMv, MV& derivedMv, uint32_t width, uint32_t height)
+{
+    const int   ctuWidth = m_slice->m_param->maxCUSize;
+    const int   ctuHeight = m_slice->m_param->maxCUSize;
+    int   cuPelX = m_cuPelX + (absPartIdx ? g_zscanToPelX[absPartIdx] : 0);
+    int   cuPelY = m_cuPelY + (absPartIdx ? g_zscanToPelX[absPartIdx] : 0);
+    int rngX = cuPelX + (currentMv.x >> 2);
+    int rngY = cuPelY + (currentMv.y >> 2);
+    uint32_t m_frameWidthInCtus = (m_slice->m_sps->picWidthInLumaSamples % ctuWidth) ? m_slice->m_sps->picWidthInLumaSamples / ctuWidth + 1 : m_slice->m_sps->picWidthInLumaSamples / ctuWidth;
+
+    if (rngX < 0 || rngY < 0 || (rngX + width) > m_slice->m_sps->picWidthInLumaSamples || (rngY + height) > m_slice->m_sps->picHeightInLumaSamples)
+    {
+        return false;
+    }
+
+    int refCtbAddr = (rngY / ctuHeight) * m_frameWidthInCtus + (rngX / ctuWidth);
+
+    int      relCUPelX = rngX & (ctuWidth - 1);
+    int      relCUPelY = rngY & (ctuHeight - 1);
+    uint32_t absPartIdxDerived = g_rasterToZscan[((relCUPelY >> 2) << 4) + (relCUPelX >> 2)];
+    CUData* refCU = m_encData->getPicCTU(refCtbAddr);
+
+    if (refCU->m_slice == NULL)
+        return false;
+
+    MVField mv1;
+    refCU->getMvField(refCU, absPartIdxDerived, 0, mv1);
+
+    int iCurrCtbAddr = (m_cuPelY / ctuHeight) * m_frameWidthInCtus + (m_cuPelX / ctuWidth);
+    uint32_t currAbsPartIdx = g_rasterToZscan[(((m_cuPelY & (ctuHeight - 1)) >> 2) << 4) + ((m_cuPelX & (ctuWidth - 1)) >> 2)];
+
+    if ((refCtbAddr > iCurrCtbAddr) || ((refCtbAddr == iCurrCtbAddr) && (absPartIdxDerived >= currAbsPartIdx)))
+        return false;
+
+    int refIdx = mv1.refIdx;
+    bool isIBC;
+    if (refCU->isIntra(absPartIdxDerived))
+    {
+        isIBC = false;
+    }
+    else
+    {
+        isIBC = (refIdx >= 0) ? (refCU->m_slice->m_refFrameList[0][refIdx]->m_poc == refCU->m_slice->m_poc) : 0;
+    }
+    derivedMv = mv1.mv;
+    derivedMv += currentMv;
+
+    return isIBC;
+
+}
+
+bool CUData::isIntraBC(const CUData* cu, uint32_t absPartIdx) const
+{
+    if (cu->isIntra(absPartIdx))
+    {
+        return false;
+    }
+    MVField mv;
+    cu->getMvField(cu, absPartIdx, 0, mv);
+    int iRefIdx = mv.refIdx;
+    bool isNeighborIntraBC = (iRefIdx >= 0) ? (m_slice->m_refFrameList[0][iRefIdx]->m_poc == m_slice->m_poc) : false;
+
+    return isNeighborIntraBC;
+}
+
+bool CUData::getColMVPIBC(int ctuRsAddr, int partUnitIdx, MV& rcMv)
+{
+    uint32_t absPartAddr = partUnitIdx;
+
+    // use coldir.
+    Frame* colPic = m_slice->m_lastEncPic;
+    if (!colPic)
+        return false;
+
+    CUData* colCU = m_encData->getPicCTU(ctuRsAddr);
+    MVField tempMv;
+    colCU->getMvField(colCU, absPartAddr, 0, tempMv);
+    if (&tempMv == NULL)
+        return false;
+
+    rcMv = tempMv.mv;
+
+    return true;
+}
+
+void CUData::getIntraBCMVPsEncOnly(uint32_t tabsPartIdx, MV* MvPred, int& nbPred, int puIdx)
+{
+    uint32_t        tempPartIdx;
+    uint32_t        left, above;
+    PartSize        ePartSize = (PartSize)m_partSize[tabsPartIdx];
+    const MVField* mvField;
+    MVField         tempMvField;
+    uint32_t        absPartIdx;
+
+    int width, height;
+    getPartIndexAndSize(puIdx, absPartIdx, width, height);
+    uint32_t            numPartInCUWidth = s_numPartInCUSize;
+    uint32_t            m_numPartitionsInCtu = s_numPartInCUSize * s_numPartInCUSize;
+    uint32_t            m_frameWidthInCtus = (m_slice->m_sps->picWidthInLumaSamples % m_slice->m_param->maxCUSize) ? m_slice->m_sps->picWidthInLumaSamples / m_slice->m_param->maxCUSize + 1 : m_slice->m_sps->picWidthInLumaSamples / m_slice->m_param->maxCUSize;
+
+    uint32_t            partIdxLT = m_absIdxInCTU + absPartIdx;
+    uint32_t            partIdxLB = g_rasterToZscan[g_zscanToRaster[partIdxLT] + (height / 4 - 1) * numPartInCUWidth];
+    uint32_t            partIdxRT = g_rasterToZscan[g_zscanToRaster[partIdxLT] + width / 4 - 1];
+
+    left = above = 0;
+
+    MvPred[0] = m_lastIntraBCMv[0];
+    if (MvPred[0] != MV(0, 0))
+    {
+        nbPred++;
+        if (getDerivedBV(absPartIdx, MvPred[nbPred - 1], MvPred[nbPred], width, height))
+            nbPred++;
+    }
+    MvPred[nbPred] = m_lastIntraBCMv[1];
+    if (MvPred[nbPred] != MV(0, 0))
+    {
+        nbPred++;
+        if (getDerivedBV(absPartIdx, MvPred[nbPred - 1], MvPred[nbPred], width, height))
+            nbPred++;
+    }
+
+    //left
+    const CUData* leftCU = getPULeft(tempPartIdx, partIdxLB);
+    left = leftCU ? isIntraBC(leftCU, tempPartIdx) : 0;
+
+    if (left)
+    {
+        leftCU->getMvField(leftCU, tempPartIdx, 0, tempMvField);
+        MvPred[nbPred++] = tempMvField.mv;
+        if (getDerivedBV(absPartIdx, MvPred[nbPred - 1], MvPred[nbPred], width, height))
+        {
+            nbPred++;
+        }
+    }
+
+    //above
+    const CUData* aboveCU = getPUAbove(tempPartIdx, partIdxRT);
+    above = aboveCU ? isIntraBC(aboveCU, tempPartIdx) : 0;
+
+    if (above)
+    {
+        aboveCU->getMvField(aboveCU, tempPartIdx, 0, tempMvField);
+        MvPred[nbPred++] = tempMvField.mv;
+        if (getDerivedBV(absPartIdx, MvPred[nbPred - 1], MvPred[nbPred], width, height))
+        {
+            nbPred++;
+        }
+    }
+
+    if (m_slice->isOnlyCurrentPictureAsReference())
+    {
+        MV mvCol;
+        bool isColAvail = false;
+        if (m_absIdxInCTU && m_slice->m_lastEncPic && m_slice->m_lastEncPic->m_poc < m_slice->m_poc)
+        {
+            uint32_t partIdxRB;
+            partIdxRB = deriveRightBottomIdx(puIdx);
+
+            uint32_t absPartIdxTmp = g_zscanToRaster[partIdxRB];
+            uint32_t absPartAddr = m_absIdxInCTU + absPartIdx;
+            int      iLCUIdx = -1;
+
+            if (((m_encData->getPicCTU(m_cuAddr)->m_cuPelX + g_zscanToPelX[g_rasterToZscan[absPartIdxTmp]] + 4) < m_slice->m_sps->picWidthInLumaSamples)  // image boundary check
+                && ((m_encData->getPicCTU(m_cuAddr)->m_cuPelY + g_zscanToPelY[g_rasterToZscan[absPartIdxTmp]] + 4) < m_slice->m_sps->picHeightInLumaSamples))
+            {
+                if ((absPartIdxTmp % numPartInCUWidth < numPartInCUWidth - 1) &&           // is not at the last column of LCU
+                    (absPartIdxTmp / numPartInCUWidth < s_numPartInCUSize - 1)) // is not at the last row    of LCU
+                {
+                    absPartAddr = g_rasterToZscan[absPartIdxTmp + numPartInCUWidth + 1];
+                    iLCUIdx = m_cuAddr;
+                }
+                else if (absPartIdxTmp % numPartInCUWidth < numPartInCUWidth - 1)           // is not at the last column of CTU But is last row of CTU
+                {
+                    absPartAddr = g_rasterToZscan[(absPartIdxTmp + numPartInCUWidth + 1) % m_numPartitionsInCtu];
+                    iLCUIdx = m_cuAddr + m_frameWidthInCtus;
+                }
+                else if (absPartIdxTmp / numPartInCUWidth < s_numPartInCUSize - 1)          // is not at the last row of CTU But is last column of CTU
+                {
+                    absPartAddr = g_rasterToZscan[absPartIdxTmp + 1];
+                    iLCUIdx = m_cuAddr + 1;
+                }
+            }
+            if (iLCUIdx >= 0)
+            {
+                isColAvail = getColMVPIBC(iLCUIdx, absPartAddr, mvCol);
+
+                if (!isColAvail)
+                {
+                    uint32_t uiPartIdxCenter;
+                    uiPartIdxCenter = deriveCenterIdx(puIdx);
+                    isColAvail = getColMVPIBC(m_cuAddr, uiPartIdxCenter, mvCol);
+                }
+            }
+        }
+        if (isColAvail)
+        {
+            MvPred[nbPred++] = mvCol;
+            if (getDerivedBV(absPartIdx, MvPred[nbPred - 1], MvPred[nbPred], width, height))
+            {
+                nbPred++;
+            }
+        }
+    }
+
+    // Below Left predictor search
+    const CUData* tempBelowLeftCU = getPUBelowLeft(tempPartIdx, partIdxLB);
+    uint32_t belowLeft = (tempBelowLeftCU) ? tempBelowLeftCU->isIntraBC(tempBelowLeftCU, tempPartIdx) : 0;
+    if (belowLeft)
+    {
+        tempBelowLeftCU->getMvField(tempBelowLeftCU, tempPartIdx, 0, tempMvField);
+        MvPred[nbPred++] = tempMvField.mv;
+        if (getDerivedBV(absPartIdx, MvPred[nbPred - 1], MvPred[nbPred], width, height))
+        {
+            nbPred++;
+        }
+    }
+
+    // Above Right predictor search
+    const CUData* tempAboveRightCU = getPUAboveRight(tempPartIdx, partIdxRT);
+    uint32_t aboveRight = (tempAboveRightCU) ? tempAboveRightCU->isIntraBC(tempAboveRightCU, tempPartIdx) : 0;
+    if (aboveRight)
+    {
+        tempAboveRightCU->getMvField(tempAboveRightCU, tempPartIdx, 0, tempMvField);
+        MvPred[nbPred++] = tempMvField.mv;
+        if (getDerivedBV(absPartIdx, MvPred[nbPred - 1], MvPred[nbPred], width, height))
+        {
+            nbPred++;
+        }
+    }
+
+    // Above Left predictor search
+    const CUData* tempAboveLeftCU = getPUAboveLeft(tempPartIdx, partIdxLT);
+    uint32_t aboveLeft = (tempAboveLeftCU) ? tempAboveLeftCU->isIntraBC(tempAboveLeftCU, tempPartIdx) : 0;
+    if (aboveLeft)
+    {
+        tempAboveLeftCU->getMvField(tempAboveLeftCU, tempPartIdx, 0, tempMvField);
+        MvPred[nbPred++] = tempMvField.mv;
+        if (getDerivedBV(absPartIdx, MvPred[nbPred - 1], MvPred[nbPred], width, height))
+        {
+            nbPred++;
+        }
+    }
+}
+
+void CUData::roundMergeCandidates(MVField(*pcMvFieldNeighbours)[2], int iCount) const
+{
+    if (0) //m_pcSlice->getUseIntegerMv()
+    {
+        for (int i = 0; i < iCount; i++)
+        {
+            pcMvFieldNeighbours[i][0].mv = (pcMvFieldNeighbours[i][0].mv >> 2) << 2;
+            pcMvFieldNeighbours[i][0].refIdx = pcMvFieldNeighbours[i][0].refIdx;
+        }
+    }
+    else
+    {
+        for (int i = 0; i < iCount; i++)
+        {
+            int iCurrRefIdx = pcMvFieldNeighbours[i][0].refIdx;
+            if (iCurrRefIdx >= 0)
+            {
+                if (m_slice->m_refFrameList[0][iCurrRefIdx]->m_poc == m_slice->m_poc)
+                {
+                    pcMvFieldNeighbours[i][0].mv = (pcMvFieldNeighbours[i][0].mv >> 2) << 2;
+                    pcMvFieldNeighbours[i][0].refIdx = pcMvFieldNeighbours[i][0].refIdx;
+                }
+            }
+        }
+    }
+}
+
+bool CUData::is8x8BipredRestriction(MV mvL0, MV mvL1, int iRefIdxL0, int iRefIdxL1) const
+{
+    if (iRefIdxL0 < -1 || iRefIdxL0 >= MAX_NUM_REF)
+    {
+        iRefIdxL0 = -1;
+    }
+    if (iRefIdxL1 < -1 || iRefIdxL1 >= MAX_NUM_REF)
+    {
+        iRefIdxL1 = -1;
+    }
+    bool b8x8BiPredRestricted = false;
+    int RefPOCL0 = -1;
+    int RefPOCL1 = -1;
+    if (iRefIdxL0 >= 0 && iRefIdxL1 >= 0)
+    {
+        RefPOCL0 = m_slice->m_refPOCList[0][iRefIdxL0];
+        RefPOCL1 = m_slice->m_refPOCList[1][iRefIdxL1];
+        bool mvL0Int = (((mvL0.x & 0x3) == 0) && ((mvL0.y & 0x3) == 0));
+        bool mvL1Int = (((mvL1.x & 0x3) == 0) && ((mvL1.y & 0x3) == 0));
+        bool IdenticalMV = ((mvL0 == mvL1) && (RefPOCL0 == RefPOCL1));
+        b8x8BiPredRestricted = (
+            !mvL0Int && !mvL1Int && !IdenticalMV &&
+            (m_slice->m_param->bEnableSCC)
+            && (m_slice->m_bUseSao || !m_slice->m_pps->bPicDisableDeblockingFilter || 0));
+    }
+    return b8x8BiPredRestricted;
+}
diff --git a/source/common/cudata.h b/source/common/cudata.h
index 404e95714..362df5786 100644
--- a/source/common/cudata.h
+++ b/source/common/cudata.h
@@ -232,6 +232,8 @@ public:
     uint32_t*       m_collectCUVariance;
     uint32_t*       m_collectCUCount;
 
+    MV              m_lastIntraBCMv[2];
+
     CUData();
 
     void     initialize(const CUDataMemPool& dataPool, uint32_t depth, const x265_param& param, int instance);
@@ -311,6 +313,13 @@ public:
     const CUData* getPUAboveRightAdi(uint32_t& arPartUnitIdx, uint32_t curPartUnitIdx, uint32_t partUnitOffset) const;
     const CUData* getPUBelowLeftAdi(uint32_t& blPartUnitIdx, uint32_t curPartUnitIdx, uint32_t partUnitOffset) const;
 
+    void getIntraBCMVPsEncOnly(uint32_t absPartIdx, MV* MvPred, int& nbPred, int puIdx);
+    bool getDerivedBV(uint32_t absPartIdx, const MV& currentMv, MV& derivedMv, uint32_t width, uint32_t height);
+    bool isIntraBC(const CUData* cu, uint32_t absPartIdx) const;
+    bool getColMVPIBC(int ctuRsAddr, int partUnitIdx, MV& rcMv);
+    void roundMergeCandidates(MVField(*candMvField)[2], int iCount) const;
+    bool is8x8BipredRestriction(MV mvL0, MV mvL1, int iRefIdxL0, int iRefIdxL1) const;
+
 protected:
 
     template<typename T>
diff --git a/source/common/slice.cpp b/source/common/slice.cpp
index badcb7671..337112e07 100644
--- a/source/common/slice.cpp
+++ b/source/common/slice.cpp
@@ -55,12 +55,28 @@ void Slice::createInterLayerReferencePictureSet(PicList& picList, PicList& refPi
 
 void Slice::setRefPicList(PicList& picList, PicList& refPicSetInterLayer0, PicList& refPicSetInterLayer1, int sLayerId)
 {
+    bool checkNumPocTotalCurr = m_param->bEnableSCC ? false : true;
     if (m_sliceType == I_SLICE)
     {
         memset(m_refFrameList, 0, sizeof(m_refFrameList));
         memset(m_refReconPicList, 0, sizeof(m_refReconPicList));
         memset(m_refPOCList, 0, sizeof(m_refPOCList));
         m_numRefIdx[1] = m_numRefIdx[0] = 0;
+
+        if (!checkNumPocTotalCurr)
+        {
+            if (m_rps.numberOfPictures == 0)
+            {
+                Frame* prevPic = picList.getPOC(max(0, m_poc - 1));
+                if (prevPic->m_poc != max(0, m_poc - 1))
+                {
+                    prevPic = picList.getPOC(m_poc);
+                }
+                m_lastEncPic = prevPic;
+            }
+            return;
+        }
+
         return;
     }
 
@@ -73,6 +89,17 @@ void Slice::setRefPicList(PicList& picList, PicList& refPicSetInterLayer0, PicLi
         m_numRefIdx[0] = 1;
     }
 
+    if (!checkNumPocTotalCurr && m_rps.numberOfPictures == 0)
+    {
+        Frame* prevPic = picList.getPOC(max(0, m_poc - 1));
+        if (prevPic->m_poc != max(0, m_poc - 1))
+        {
+            prevPic = picList.getPOC(m_poc);
+
+        }
+        m_lastEncPic = prevPic;
+    }
+
     Frame* refPic = NULL;
     Frame* refPicSetStCurr0[MAX_NUM_REF];
     Frame* refPicSetStCurr1[MAX_NUM_REF];
@@ -226,6 +253,32 @@ void Slice::disableWeights()
             }
 }
 
+bool Slice::isOnlyCurrentPictureAsReference() const
+{
+    if (m_sliceType == I_SLICE)
+    {
+        return true;
+    }
+
+    for (int i = 0; i < m_numRefIdx[0]; i++)
+    {
+        if (m_refFrameList[0][i]->m_poc != m_poc)
+        {
+            return false;
+        }
+    }
+
+    for (int i = 0; i < m_numRefIdx[1]; i++)
+    {
+        if (m_refFrameList[1][i]->m_poc != m_poc)
+        {
+            return false;
+        }
+    }
+
+    return true;
+}
+
 /* Sorts the deltaPOC and Used by current values in the RPS based on the
  * deltaPOC values.  deltaPOC values are sorted with -ve values before the +ve
  * values.  -ve values are in decreasing order.  +ve values are in increasing
diff --git a/source/common/slice.h b/source/common/slice.h
index f03508660..2f8872473 100644
--- a/source/common/slice.h
+++ b/source/common/slice.h
@@ -403,6 +403,9 @@ public:
     int         m_fieldNum;
     Frame*      m_mcstfRefFrameList[2][MAX_MCSTF_TEMPORAL_WINDOW_LENGTH];
 
+    Frame*      m_lastEncPic;
+    bool        m_bLMvdL1Zero;
+
     Slice()
     {
         m_lastIDR = 0;
@@ -427,6 +430,8 @@ public:
     void createInterLayerReferencePictureSet(PicList& picList, PicList& refPicSetInterLayer0, PicList& refPicSetInterLayer1);
 #endif
 
+    bool isOnlyCurrentPictureAsReference() const;
+
     bool getRapPicFlag() const
     {
         return m_nalUnitType == NAL_UNIT_CODED_SLICE_IDR_W_RADL
diff --git a/source/encoder/analysis.cpp b/source/encoder/analysis.cpp
index 33bf9d9a9..4bbfcbfe3 100644
--- a/source/encoder/analysis.cpp
+++ b/source/encoder/analysis.cpp
@@ -517,6 +517,18 @@ uint64_t Analysis::compressIntraCU(const CUData& parentCTU, const CUGeom& cuGeom
     ModeDepth& md = m_modeDepth[depth];
     md.bestMode = NULL;
 
+    MV lastIntraBCMv[2];
+    for (int i = 0; i < 2; i++)
+    {
+        if (depth == 0)
+            lastIntraBCMv[i] = parentCTU.m_lastIntraBCMv[i];
+        else
+            lastIntraBCMv[i] = tempLastIntraBCMv[i];
+    }
+
+    MV iMVCandList[4][10];
+    memset(iMVCandList, 0, sizeof(MV) * 4 * 10);
+
     bool mightSplit = !(cuGeom.flags & CUGeom::LEAF);
     bool mightNotSplit = !(cuGeom.flags & CUGeom::SPLIT_MANDATORY);
 
@@ -567,6 +579,54 @@ uint64_t Analysis::compressIntraCU(const CUData& parentCTU, const CUGeom& cuGeom
             checkBestMode(md.pred[PRED_INTRA_NxN], depth);
         }
 
+        bool intraBlockCopyFastSearch = (m_param->bEnableSCC == 1) ? true : false, bSkipIntraBlockCopySearch = false, bUse1DSearchFor8x8 = true;
+        uint32_t cost;
+        double intracost = MAX_DOUBLE;
+        if (m_param->bEnableSCC)
+        {
+            md.pred[PRED_MERGE_IBC].cu.initSubCU(parentCTU, cuGeom, qp);
+            checkRDCostIntraBCMerge2Nx2N(md.pred[PRED_MERGE_IBC], cuGeom);
+
+            if (!bSkipIntraBlockCopySearch)
+            {
+                md.pred[PRED_IBC_2Nx2N].cu.initSubCU(parentCTU, cuGeom, qp);
+                checkIntraBC_rd5_6(md.pred[PRED_IBC_2Nx2N], cuGeom, SIZE_2Nx2N, false, bUse1DSearchFor8x8, tempLastIntraBCMv, &MV(0, 0));
+                checkBestMode(md.pred[PRED_IBC_2Nx2N], depth);
+
+                if (intraBlockCopyFastSearch)
+                {
+                    if (depth == m_slice->m_sps->log2DiffMaxMinCodingBlockSize)
+                    {
+                        intracost = min(intracost, md.pred[PRED_IBC_2Nx2N].rdCost);
+                        double dTH2 = max(60 * m_rdCost.m_lambda, 56.0);
+                        double dTH3 = max(66 * m_rdCost.m_lambda, 800.0);
+
+
+                        md.pred[PRED_IBC_Nx2N].cu.initSubCU(parentCTU, cuGeom, qp);
+                        checkIntraBC_rd5_6(md.pred[PRED_IBC_Nx2N], cuGeom, SIZE_Nx2N, false, bUse1DSearchFor8x8, tempLastIntraBCMv, &MV(0, 0));
+                        checkBestMode(md.pred[PRED_IBC_Nx2N], depth);
+                        intracost = min(intracost, md.pred[PRED_IBC_Nx2N].rdCost);
+
+                        md.pred[PRED_IBC_2NxN].cu.initSubCU(parentCTU, cuGeom, qp);
+                        checkIntraBC_rd5_6(md.pred[PRED_IBC_2NxN], cuGeom, SIZE_2NxN, false, bUse1DSearchFor8x8, tempLastIntraBCMv, &MV(0, 0));
+                        checkBestMode(md.pred[PRED_IBC_2NxN], depth);
+                        intracost = min(intracost, md.pred[PRED_IBC_2NxN].rdCost);
+                    }
+                }
+                else
+                {
+                    md.pred[PRED_IBC_2NxN].cu.initSubCU(parentCTU, cuGeom, qp);
+                    checkIntraBC_rd5_6(md.pred[PRED_IBC_2NxN], cuGeom, SIZE_2NxN, false, bUse1DSearchFor8x8, tempLastIntraBCMv, &MV(0, 0));
+                    checkBestMode(md.pred[PRED_IBC_2NxN], depth);
+
+                    md.pred[PRED_IBC_Nx2N].cu.initSubCU(parentCTU, cuGeom, qp);
+                    checkIntraBC_rd5_6(md.pred[PRED_IBC_Nx2N], cuGeom, SIZE_Nx2N, false, bUse1DSearchFor8x8, tempLastIntraBCMv, &MV(0, 0));
+                    checkBestMode(md.pred[PRED_IBC_Nx2N], depth);
+                }
+            }
+
+        }
+
         if (m_bTryLossless)
             tryLossless(cuGeom);
 
@@ -574,6 +634,89 @@ uint64_t Analysis::compressIntraCU(const CUData& parentCTU, const CUGeom& cuGeom
             addSplitFlagCost(*md.bestMode, cuGeom.depth);
     }
 
+    // If Intra BC keep last coded Mv
+    if (md.bestMode && md.bestMode->cu.isInter(0))
+    {
+        MVField mvField;
+        const CUData* cu = &md.bestMode->cu;
+        md.bestMode->cu.getMvField(cu, 0, 0, mvField);
+        int iRefIdxFirst = mvField.refIdx;
+        md.bestMode->cu.getMvField(cu, cu->m_numPartitions - 1, 0, mvField);
+        int iRefIdxLast = mvField.refIdx;
+        bool isIntraBCFirst = (iRefIdxFirst >= 0) ? cu->m_slice->m_refFrameList[0][iRefIdxFirst]->m_poc == cu->m_slice->m_poc : false;
+        bool isIntraBCLast = (iRefIdxLast >= 0) ? cu->m_slice->m_refFrameList[0][iRefIdxLast]->m_poc == cu->m_slice->m_poc : false;
+
+        if (isIntraBCFirst || isIntraBCLast)
+        {
+            if (cu->m_partSize[0] == SIZE_2Nx2N)
+            {
+                md.bestMode->cu.getMvField(cu, cu->m_numPartitions - 1, 0, mvField);
+                if (mvField.mv != cu->m_lastIntraBCMv[0])
+                {
+                    md.bestMode->cu.m_lastIntraBCMv[1] = cu->m_lastIntraBCMv[0];
+                    md.bestMode->cu.m_lastIntraBCMv[0] = mvField.mv;
+                }
+            }
+            else if (cu->m_partSize[0] == SIZE_2NxN || cu->m_partSize[0] == SIZE_Nx2N)
+            {
+                // mixed PU, only one partition is IntraBC coded
+                if (isIntraBCFirst != isIntraBCLast)
+                {
+                    if (isIntraBCFirst)
+                    {
+                        // Part 0
+                        md.bestMode->cu.getMvField(cu, 0, 0, mvField);
+                        if (mvField.mv != cu->m_lastIntraBCMv[0])
+                        {
+                            md.bestMode->cu.m_lastIntraBCMv[1] = cu->m_lastIntraBCMv[0];
+                            md.bestMode->cu.m_lastIntraBCMv[0] = mvField.mv;
+                        }
+                    }
+                    else if (isIntraBCLast)
+                    {
+                        // Part 1
+                        md.bestMode->cu.getMvField(cu, cu->m_numPartitions - 1, 0, mvField);
+                        if (mvField.mv != cu->m_lastIntraBCMv[0])
+                        {
+                            md.bestMode->cu.m_lastIntraBCMv[1] = cu->m_lastIntraBCMv[0];
+                            md.bestMode->cu.m_lastIntraBCMv[0] = mvField.mv;
+                        }
+                    }
+                }
+                else // normal IntraBC CU
+                {
+                    // Part 0
+                    md.bestMode->cu.getMvField(cu, 0, 0, mvField);
+                    if (mvField.mv != cu->m_lastIntraBCMv[0])
+                    {
+                        md.bestMode->cu.m_lastIntraBCMv[1] = cu->m_lastIntraBCMv[0];
+                        md.bestMode->cu.m_lastIntraBCMv[0] = mvField.mv;
+                    }
+                    // Part 1
+                    md.bestMode->cu.getMvField(cu, cu->m_numPartitions - 1, 0, mvField);
+                    if (mvField.mv != cu->m_lastIntraBCMv[0])
+                    {
+                        md.bestMode->cu.m_lastIntraBCMv[1] = cu->m_lastIntraBCMv[0];
+                        md.bestMode->cu.m_lastIntraBCMv[0] = mvField.mv;
+                    }
+                }
+            }
+            else
+            {
+                // NxN
+                for (int part = 0; part < 4; part++)
+                {
+                    md.bestMode->cu.getMvField(cu, cu->m_numPartitions - 4 + part, 0, mvField);
+                    if (mvField.mv != cu->m_lastIntraBCMv[0])
+                    {
+                        md.bestMode->cu.m_lastIntraBCMv[1] = cu->m_lastIntraBCMv[0];
+                        md.bestMode->cu.m_lastIntraBCMv[0] = mvField.mv;
+                    }
+                }
+            }
+        }
+    } // is inter
+
     // stop recursion if we reach the depth of previous analysis decision
     mightSplit &= !(bAlreadyDecided && bDecidedDepth) || split;
 
@@ -603,6 +746,9 @@ uint64_t Analysis::compressIntraCU(const CUData& parentCTU, const CUGeom& cuGeom
                 if (m_slice->m_pps->bUseDQP && nextDepth <= m_slice->m_pps->maxCuDQPDepth)
                     nextQP = setLambdaFromQP(parentCTU, calculateQpforCuSize(parentCTU, childGeom));
 
+                for (int i = 0; i < 2; i++)
+                    tempLastIntraBCMv[i] = lastIntraBCMv[i];
+
                 if (m_param->bEnableSplitRdSkip)
                 {
                     curCost += compressIntraCU(parentCTU, childGeom, nextQP);
@@ -615,6 +761,14 @@ uint64_t Analysis::compressIntraCU(const CUData& parentCTU, const CUGeom& cuGeom
                 else
                     compressIntraCU(parentCTU, childGeom, nextQP);
 
+                if (nd.bestMode->cu.m_lastIntraBCMv[0].x != 0 || nd.bestMode->cu.m_lastIntraBCMv[0].y != 0)
+                {
+                    for (int i = 0; i < 2; i++)
+                    {
+                        lastIntraBCMv[i] = nd.bestMode->cu.m_lastIntraBCMv[i];
+                    }
+                }
+
                 // Save best CU and pred data for this sub CU
                 splitCU->copyPartFrom(nd.bestMode->cu, childGeom, subPartIdx);
                 splitPred->addSubCosts(*nd.bestMode);
@@ -1856,6 +2010,19 @@ SplitData Analysis::compressInterCU_rd5_6(const CUData& parentCTU, const CUGeom&
     ModeDepth& md = m_modeDepth[depth];
     md.bestMode = NULL;
 
+    Mode* interBest; // store the best modes in inter prediction
+    MV lastIntraBCMv[2];
+    for (int i = 0; i < 2; i++)
+    {
+        if (depth == 0)
+            lastIntraBCMv[i] = parentCTU.m_lastIntraBCMv[i];
+        else
+            lastIntraBCMv[i] = tempLastIntraBCMv[i];
+    }
+
+    MV iMVCandList[4][10];
+    memset(iMVCandList, 0, sizeof(MV) * 4 * 10);
+
     if (m_param->searchMethod == X265_SEA)
     {
         int numPredDir = m_slice->isInterP() ? 1 : 2;
@@ -2023,6 +2190,13 @@ SplitData Analysis::compressInterCU_rd5_6(const CUData& parentCTU, const CUGeom&
             checkInter_rd5_6(md.pred[PRED_2Nx2N], cuGeom, SIZE_2Nx2N, refMasks);
             checkBestMode(md.pred[PRED_2Nx2N], cuGeom.depth);
 
+            interBest = md.bestMode;
+            if (m_param->bEnableSCC)
+            {
+                md.pred[PRED_MERGE_IBC].cu.initSubCU(parentCTU, cuGeom, qp);
+                checkRDCostIntraBCMerge2Nx2N(md.pred[PRED_MERGE_IBC], cuGeom);
+            }
+
             if (m_param->recursionSkipMode == RDCOST_BASED_RSKIP && depth && m_modeDepth[depth - 1].bestMode)
                 skipRecursion = md.bestMode && !md.bestMode->cu.getQtRootCbf(0);
             else if (cuGeom.log2CUSize >= MAX_LOG2_CU_SIZE - 1 && m_param->recursionSkipMode == EDGE_BASED_RSKIP)
@@ -2059,8 +2233,17 @@ SplitData Analysis::compressInterCU_rd5_6(const CUData& parentCTU, const CUGeom&
                     if (m_slice->m_pps->bUseDQP && nextDepth <= m_slice->m_pps->maxCuDQPDepth)
                         nextQP = setLambdaFromQP(parentCTU, calculateQpforCuSize(parentCTU, childGeom));
 
+                    for (int i = 0; i < 2; i++)
+                        tempLastIntraBCMv[i] = lastIntraBCMv[i];
+
                     splitData[subPartIdx] = compressInterCU_rd5_6(parentCTU, childGeom, nextQP);
 
+                    if (nd.bestMode->cu.m_lastIntraBCMv[0].x != 0 || nd.bestMode->cu.m_lastIntraBCMv[0].y != 0)
+                    {
+                        for (int i = 0; i < 2; i++)
+                            lastIntraBCMv[i] = nd.bestMode->cu.m_lastIntraBCMv[i];
+                    }
+
                     // Save best CU and pred data for this sub CU
                     splitIntra |= nd.bestMode->cu.isIntra(0);
                     splitCU->copyPartFrom(nd.bestMode->cu, childGeom, subPartIdx);
@@ -2162,6 +2345,7 @@ SplitData Analysis::compressInterCU_rd5_6(const CUData& parentCTU, const CUGeom&
                             md.pred[PRED_2NxN].cu.initSubCU(parentCTU, cuGeom, qp);
                             checkInter_rd5_6(md.pred[PRED_2NxN], cuGeom, SIZE_2NxN, refMasks);
                             checkBestMode(md.pred[PRED_2NxN], cuGeom.depth);
+                            interBest = (md.pred[PRED_2NxN].rdCost < interBest->rdCost) ? &md.pred[PRED_2NxN] : interBest;
                         }
 
                         if (splitCost < md.bestMode->rdCost + threshold_Nx2N)
@@ -2171,6 +2355,7 @@ SplitData Analysis::compressInterCU_rd5_6(const CUData& parentCTU, const CUGeom&
                             md.pred[PRED_Nx2N].cu.initSubCU(parentCTU, cuGeom, qp);
                             checkInter_rd5_6(md.pred[PRED_Nx2N], cuGeom, SIZE_Nx2N, refMasks);
                             checkBestMode(md.pred[PRED_Nx2N], cuGeom.depth);
+                            interBest = (md.pred[PRED_Nx2N].rdCost < interBest->rdCost) ? &md.pred[PRED_Nx2N] : interBest;
                         }
 
                         if (!try_2NxN_first && splitCost < md.bestMode->rdCost + threshold_2NxN)
@@ -2180,6 +2365,7 @@ SplitData Analysis::compressInterCU_rd5_6(const CUData& parentCTU, const CUGeom&
                             md.pred[PRED_2NxN].cu.initSubCU(parentCTU, cuGeom, qp);
                             checkInter_rd5_6(md.pred[PRED_2NxN], cuGeom, SIZE_2NxN, refMasks);
                             checkBestMode(md.pred[PRED_2NxN], cuGeom.depth);
+                            interBest = (md.pred[PRED_2NxN].rdCost < interBest->rdCost) ? &md.pred[PRED_2NxN] : interBest;
                         }
                     }
 
@@ -2231,6 +2417,7 @@ SplitData Analysis::compressInterCU_rd5_6(const CUData& parentCTU, const CUGeom&
                                 md.pred[PRED_2NxnD].cu.initSubCU(parentCTU, cuGeom, qp);
                                 checkInter_rd5_6(md.pred[PRED_2NxnD], cuGeom, SIZE_2NxnD, refMasks);
                                 checkBestMode(md.pred[PRED_2NxnD], cuGeom.depth);
+                                interBest = (md.pred[PRED_2NxnD].rdCost < interBest->rdCost) ? &md.pred[PRED_2NxnD] : interBest;
                             }
 
                             if (splitCost < md.bestMode->rdCost + threshold_2NxnU)
@@ -2240,6 +2427,7 @@ SplitData Analysis::compressInterCU_rd5_6(const CUData& parentCTU, const CUGeom&
                                 md.pred[PRED_2NxnU].cu.initSubCU(parentCTU, cuGeom, qp);
                                 checkInter_rd5_6(md.pred[PRED_2NxnU], cuGeom, SIZE_2NxnU, refMasks);
                                 checkBestMode(md.pred[PRED_2NxnU], cuGeom.depth);
+                                interBest = (md.pred[PRED_2NxnU].rdCost < interBest->rdCost) ? &md.pred[PRED_2NxnU] : interBest;
                             }
 
                             if (!try_2NxnD_first && splitCost < md.bestMode->rdCost + threshold_2NxnD)
@@ -2249,6 +2437,7 @@ SplitData Analysis::compressInterCU_rd5_6(const CUData& parentCTU, const CUGeom&
                                 md.pred[PRED_2NxnD].cu.initSubCU(parentCTU, cuGeom, qp);
                                 checkInter_rd5_6(md.pred[PRED_2NxnD], cuGeom, SIZE_2NxnD, refMasks);
                                 checkBestMode(md.pred[PRED_2NxnD], cuGeom.depth);
+                                interBest = (md.pred[PRED_2NxnD].rdCost < interBest->rdCost) ? &md.pred[PRED_2NxnD] : interBest;
                             }
                         }
 
@@ -2262,6 +2451,7 @@ SplitData Analysis::compressInterCU_rd5_6(const CUData& parentCTU, const CUGeom&
                                 md.pred[PRED_nRx2N].cu.initSubCU(parentCTU, cuGeom, qp);
                                 checkInter_rd5_6(md.pred[PRED_nRx2N], cuGeom, SIZE_nRx2N, refMasks);
                                 checkBestMode(md.pred[PRED_nRx2N], cuGeom.depth);
+                                interBest = (md.pred[PRED_nRx2N].rdCost < interBest->rdCost) ? &md.pred[PRED_nRx2N] : interBest;
                             }
 
                             if (splitCost < md.bestMode->rdCost + threshold_nLx2N)
@@ -2271,6 +2461,7 @@ SplitData Analysis::compressInterCU_rd5_6(const CUData& parentCTU, const CUGeom&
                                 md.pred[PRED_nLx2N].cu.initSubCU(parentCTU, cuGeom, qp);
                                 checkInter_rd5_6(md.pred[PRED_nLx2N], cuGeom, SIZE_nLx2N, refMasks);
                                 checkBestMode(md.pred[PRED_nLx2N], cuGeom.depth);
+                                interBest = (md.pred[PRED_nLx2N].rdCost < interBest->rdCost) ? &md.pred[PRED_nLx2N] : interBest;
                             }
 
                             if (!try_nRx2N_first && splitCost < md.bestMode->rdCost + threshold_nRx2N)
@@ -2280,8 +2471,69 @@ SplitData Analysis::compressInterCU_rd5_6(const CUData& parentCTU, const CUGeom&
                                 md.pred[PRED_nRx2N].cu.initSubCU(parentCTU, cuGeom, qp);
                                 checkInter_rd5_6(md.pred[PRED_nRx2N], cuGeom, SIZE_nRx2N, refMasks);
                                 checkBestMode(md.pred[PRED_nRx2N], cuGeom.depth);
+                                interBest = (md.pred[PRED_nRx2N].rdCost < interBest->rdCost) ? &md.pred[PRED_nRx2N] : interBest;
+                            }
+                        }
+                    }
+                }
+
+                if (m_param->bEnableSCC)
+                {
+                    interBest = md.bestMode;
+                    bool intraBlockCopyFastSearch = (m_param->bEnableSCC == 1) ? true : false, bSkipIntraBlockCopySearch = false, bUse1DSearchFor8x8 = false, bValid;
+                    double intracost = MAX_DOUBLE;
+
+                    if (!bSkipIntraBlockCopySearch)
+                    {
+                        md.pred[PRED_IBC_2Nx2N].cu.initSubCU(parentCTU, cuGeom, qp);
+                        checkIntraBC_rd5_6(md.pred[PRED_IBC_2Nx2N], cuGeom, SIZE_2Nx2N, false, bUse1DSearchFor8x8, tempLastIntraBCMv, &MV(0, 0));
+                        checkBestMode(md.pred[PRED_IBC_2Nx2N], depth);
+
+                        if (intraBlockCopyFastSearch)
+                        {
+                            if (depth == m_slice->m_sps->log2DiffMaxMinCodingBlockSize)
+                            {
+                                intracost = min(intracost, md.pred[PRED_IBC_2Nx2N].rdCost);
+                                double dTH2 = max(60 * m_rdCost.m_lambda, 56.0);
+                                double dTH3 = max(66 * m_rdCost.m_lambda, 800.0);
+
+                                md.pred[PRED_IBC_Nx2N].cu.initSubCU(parentCTU, cuGeom, qp);
+                                checkIntraBC_rd5_6(md.pred[PRED_IBC_Nx2N], cuGeom, SIZE_Nx2N, false, bUse1DSearchFor8x8, tempLastIntraBCMv, (iMVCandList[SIZE_Nx2N] + 8));
+                                checkBestMode(md.pred[PRED_IBC_Nx2N], depth);
+                                intracost = min(intracost, md.pred[PRED_IBC_Nx2N].rdCost);
+
+                                md.pred[PRED_MIXED_IBC_NX2N].cu.initSubCU(parentCTU, cuGeom, qp);
+                                bValid = predMixedIntraBCInterSearch(md.pred[PRED_MIXED_IBC_NX2N], cuGeom, m_csp != X265_CSP_I400 && m_frame->m_fencPic->m_picCsp != X265_CSP_I400, SIZE_Nx2N, false, iMVCandList[SIZE_Nx2N]);
+                                if (bValid)
+                                    encodeResAndCalcRdInterCU(md.pred[PRED_MIXED_IBC_NX2N], cuGeom);
+                                else
+                                    md.pred[PRED_MIXED_IBC_NX2N].rdCost = UINT64_MAX;
+                                checkBestMode(md.pred[PRED_MIXED_IBC_NX2N], depth);
+
+                                md.pred[PRED_IBC_2NxN].cu.initSubCU(parentCTU, cuGeom, qp);
+                                checkIntraBC_rd5_6(md.pred[PRED_IBC_2NxN], cuGeom, SIZE_2NxN, false, bUse1DSearchFor8x8, tempLastIntraBCMv, (iMVCandList[SIZE_2NxN] + 8));
+                                checkBestMode(md.pred[PRED_IBC_2NxN], depth);
+                                intracost = min(intracost, md.pred[PRED_IBC_2NxN].rdCost);
+
+                                md.pred[PRED_MIXED_IBC_2NXN].cu.initSubCU(parentCTU, cuGeom, qp);
+                                bValid = predMixedIntraBCInterSearch(md.pred[PRED_MIXED_IBC_2NXN], cuGeom, m_csp != X265_CSP_I400 && m_frame->m_fencPic->m_picCsp != X265_CSP_I400, SIZE_2NxN, false, iMVCandList[SIZE_2NxN]);
+                                if (bValid)
+                                    encodeResAndCalcRdInterCU(md.pred[PRED_MIXED_IBC_2NXN], cuGeom);
+                                else
+                                    md.pred[PRED_MIXED_IBC_2NXN].rdCost = UINT64_MAX;
+                                checkBestMode(md.pred[PRED_MIXED_IBC_2NXN], depth);
                             }
                         }
+                        else // full search
+                        {
+                            md.pred[PRED_IBC_2NxN].cu.initSubCU(parentCTU, cuGeom, qp);
+                            checkIntraBC_rd5_6(md.pred[PRED_IBC_2NxN], cuGeom, SIZE_2NxN, false, bUse1DSearchFor8x8, tempLastIntraBCMv, &MV(0, 0));
+                            checkBestMode(md.pred[PRED_IBC_2NxN], depth);
+
+                            md.pred[PRED_IBC_Nx2N].cu.initSubCU(parentCTU, cuGeom, qp);
+                            checkIntraBC_rd5_6(md.pred[PRED_IBC_Nx2N], cuGeom, SIZE_Nx2N, false, bUse1DSearchFor8x8, tempLastIntraBCMv, &MV(0, 0));
+                            checkBestMode(md.pred[PRED_IBC_Nx2N], depth);
+                        }
                     }
                 }
 
@@ -2308,6 +2560,89 @@ SplitData Analysis::compressInterCU_rd5_6(const CUData& parentCTU, const CUGeom&
                 }
             }
 
+            // If Intra BC keep last coded Mv
+            if (md.bestMode->cu.isInter(0))
+            {
+                MVField mvField;
+                const CUData* cu = &md.bestMode->cu;
+                md.bestMode->cu.getMvField(cu, 0, 0, mvField);
+                int iRefIdxFirst = mvField.refIdx;
+                md.bestMode->cu.getMvField(cu, cu->m_numPartitions - 1, 0, mvField);
+                int iRefIdxLast = mvField.refIdx;
+                bool isIntraBCFirst = (iRefIdxFirst >= 0) ? cu->m_slice->m_refFrameList[0][iRefIdxFirst]->m_poc == cu->m_slice->m_poc : false;
+                bool isIntraBCLast = (iRefIdxLast >= 0) ? cu->m_slice->m_refFrameList[0][iRefIdxLast]->m_poc == cu->m_slice->m_poc : false;
+
+                if (isIntraBCFirst || isIntraBCLast)
+                {
+                    if (cu->m_partSize[0] == SIZE_2Nx2N)
+                    {
+                        md.bestMode->cu.getMvField(cu, cu->m_numPartitions - 1, 0, mvField);
+                        if (mvField.mv != cu->m_lastIntraBCMv[0])
+                        {
+                            md.bestMode->cu.m_lastIntraBCMv[1] = cu->m_lastIntraBCMv[0];
+                            md.bestMode->cu.m_lastIntraBCMv[0] = mvField.mv;
+                        }
+                    }
+                    else if (cu->m_partSize[0] == SIZE_2NxN || cu->m_partSize[0] == SIZE_Nx2N)
+                    {
+                        // mixed PU, only one partition is IntraBC coded
+                        if (isIntraBCFirst != isIntraBCLast)
+                        {
+                            if (isIntraBCFirst)
+                            {
+                                // Part 0
+                                md.bestMode->cu.getMvField(cu, 0, 0, mvField);
+                                if (mvField.mv != cu->m_lastIntraBCMv[0])
+                                {
+                                    md.bestMode->cu.m_lastIntraBCMv[1] = cu->m_lastIntraBCMv[0];
+                                    md.bestMode->cu.m_lastIntraBCMv[0] = mvField.mv;
+                                }
+                            }
+                            else if (isIntraBCLast)
+                            {
+                                // Part 1
+                                md.bestMode->cu.getMvField(cu, cu->m_numPartitions - 1, 0, mvField);
+                                if (mvField.mv != cu->m_lastIntraBCMv[0])
+                                {
+                                    md.bestMode->cu.m_lastIntraBCMv[1] = cu->m_lastIntraBCMv[0];
+                                    md.bestMode->cu.m_lastIntraBCMv[0] = mvField.mv;
+                                }
+                            }
+                        }
+                        else // normal IntraBC CU
+                        {
+                            // Part 0
+                            md.bestMode->cu.getMvField(cu, 0, 0, mvField);
+                            if (mvField.mv != cu->m_lastIntraBCMv[0])
+                            {
+                                md.bestMode->cu.m_lastIntraBCMv[1] = cu->m_lastIntraBCMv[0];
+                                md.bestMode->cu.m_lastIntraBCMv[0] = mvField.mv;
+                            }
+                            // Part 1
+                            md.bestMode->cu.getMvField(cu, cu->m_numPartitions - 1, 0, mvField);
+                            if (mvField.mv != cu->m_lastIntraBCMv[0])
+                            {
+                                md.bestMode->cu.m_lastIntraBCMv[1] = cu->m_lastIntraBCMv[0];
+                                md.bestMode->cu.m_lastIntraBCMv[0] = mvField.mv;
+                            }
+                        }
+                    }
+                    else
+                    {
+                        // NxN
+                        for (int part = 0; part < 4; part++)
+                        {
+                            md.bestMode->cu.getMvField(cu, cu->m_numPartitions - 4 + part, 0, mvField);
+                            if (mvField.mv != cu->m_lastIntraBCMv[0])
+                            {
+                                md.bestMode->cu.m_lastIntraBCMv[1] = cu->m_lastIntraBCMv[0];
+                                md.bestMode->cu.m_lastIntraBCMv[0] = mvField.mv;
+                            }
+                        }
+                    }
+                }
+            } // is inter
+
             if ((md.bestMode->cu.isInter(0) && !(md.bestMode->cu.m_mergeFlag[0] && md.bestMode->cu.m_partSize[0] == SIZE_2Nx2N)) && (m_frame->m_fencPic->m_picCsp == X265_CSP_I400 && m_csp != X265_CSP_I400))
             {
                 uint32_t numPU = md.bestMode->cu.getNumPartInter(0);
@@ -2357,7 +2692,7 @@ SplitData Analysis::compressInterCU_rd5_6(const CUData& parentCTU, const CUGeom&
             else
             {
                 /* use best merge/inter mode, in case of intra use 2Nx2N inter references */
-                CUData& cu = md.bestMode->cu.isIntra(0) ? md.pred[PRED_2Nx2N].cu : md.bestMode->cu;
+                CUData& cu = md.bestMode->cu.isIntra(0) ? md.pred[PRED_2Nx2N].cu : interBest->cu;
                 uint32_t numPU = cu.getNumPartInter(0);
                 for (uint32_t puIdx = 0, subPartIdx = 0; puIdx < numPU; puIdx++, subPartIdx += cu.getPUOffset(puIdx, 0))
                     splitCUData.splitRefs |= cu.getBestRefIdx(subPartIdx);
@@ -2902,6 +3237,8 @@ void Analysis::checkMerge2Nx2N_rd5_6(Mode& skip, Mode& merge, const CUGeom& cuGe
     MVField candMvField[MRG_MAX_NUM_CANDS][2]; // double length for mv of both lists
     uint8_t candDir[MRG_MAX_NUM_CANDS];
     uint32_t numMergeCand = merge.cu.getInterMergeCandidates(0, 0, candMvField, candDir);
+    restrictBipredMergeCand(&merge.cu, 0, candMvField, candDir, numMergeCand);
+
     PredictionUnit pu(merge.cu, cuGeom, 0);
 
     bool foundCbf0Merge = false;
@@ -2956,6 +3293,10 @@ void Analysis::checkMerge2Nx2N_rd5_6(Mode& skip, Mode& merge, const CUGeom& cuGe
             candMvField[i][0].mv.x > maxSafeMv)
             // skip merge candidates which reference beyond safe reference area
             continue;
+        if ((candDir[i] == 1 || candDir[i] == 3) && (m_slice->m_refPOCList[0][candMvField[i][0].refIdx] == m_slice->m_poc))
+        {
+            continue;
+        }
         tempPred->cu.m_mvpIdx[0][0] = (uint8_t)i;    /* merge candidate ID is stored in L0 MVP idx */
         tempPred->cu.m_interDir[0] = candDir[i];
         tempPred->cu.m_mv[0][0] = candMvField[i][0].mv;
@@ -3020,6 +3361,73 @@ void Analysis::checkMerge2Nx2N_rd5_6(Mode& skip, Mode& merge, const CUGeom& cuGe
     }
 }
 
+void Analysis::checkRDCostIntraBCMerge2Nx2N(Mode& mergeIBC, const CUGeom& cuGeom)
+{
+    mergeIBC.initCosts();
+    MVField  cMvFieldNeighbours[MRG_MAX_NUM_CANDS][2]; // double length for mv of both lists
+    uint8_t interDirNeighbours[MRG_MAX_NUM_CANDS];
+    uint32_t numValidMergeCand = 0;
+    CUData cu = mergeIBC.cu;
+    PredictionUnit pu(mergeIBC.cu, cuGeom, 0);
+    mergeIBC.rdCost = MAX_INT64;
+    for (uint32_t ui = 0; ui < m_slice->m_maxNumMergeCand; ++ui)
+    {
+        interDirNeighbours[ui] = 0;
+    }
+    int org_qp;
+    int xPos = cu.m_cuPelX;
+    int yPos = cu.m_cuPelY;
+    int width = 1 << cu.m_log2CUSize[0];
+    int height = 1 << cu.m_log2CUSize[0];
+    uint8_t depth = cu.m_cuDepth[0];
+    mergeIBC.cu.setPartSizeSubParts(SIZE_2Nx2N);
+    Mode tempPred = m_modeDepth[depth].pred[PRED_MERGE_IBC];
+
+    numValidMergeCand = cu.getInterMergeCandidates(0, 0, cMvFieldNeighbours, interDirNeighbours);
+    cu.roundMergeCandidates(cMvFieldNeighbours, numValidMergeCand);
+    restrictBipredMergeCand(&cu, 0, cMvFieldNeighbours, interDirNeighbours, numValidMergeCand);
+
+    for (uint32_t mergeCand = 0; mergeCand < numValidMergeCand; ++mergeCand)
+    {
+        if (interDirNeighbours[mergeCand] != 1)
+        {
+            continue;
+        }
+
+        if (m_slice->m_refPOCList[0][cMvFieldNeighbours[mergeCand][0].refIdx] != m_slice->m_poc)
+        {
+            continue;
+        }
+
+        if (!isBlockVectorValid(xPos, yPos, width, height, &cu,
+            0, 0, (cMvFieldNeighbours[mergeCand][0].mv.x >> 2), (cMvFieldNeighbours[mergeCand][0].mv.y >> 2), m_param->maxCUSize))
+        {
+            continue;
+        }
+
+        // set MC parameters
+        cu.setPredModeSubParts(MODE_INTER);
+        cu.setPartSizeSubParts(SIZE_2Nx2N);
+        cu.m_mergeFlag[0] = true;
+        cu.m_mvpIdx[0][0] = mergeCand;
+        cu.setPUInterDir(interDirNeighbours[mergeCand], 0, 0);
+        cu.setPUMv(0, cMvFieldNeighbours[mergeCand][0].mv, 0, 0);
+        cu.setPUMv(1, cMvFieldNeighbours[mergeCand][1].mv, 0, 0);
+        cu.setPURefIdx(0, (int8_t)cMvFieldNeighbours[mergeCand][0].refIdx, 0, 0);
+        cu.setPURefIdx(1, (int8_t)cMvFieldNeighbours[mergeCand][1].refIdx, 0, 0);
+        motionCompensation(cu, pu, mergeIBC.predYuv, true, m_csp != X265_CSP_I400);
+
+        org_qp = cu.m_qp[0];
+        encodeResAndCalcRdInterCU(mergeIBC, cuGeom);
+        if (mergeIBC.rdCost < tempPred.rdCost)
+            std::swap(mergeIBC, tempPred);
+        cu.setQPSubParts(org_qp, 0, depth);
+    }
+    std::swap(tempPred, mergeIBC);
+    checkBestMode(mergeIBC, depth);
+    checkDQP(mergeIBC, cuGeom);
+}
+
 void Analysis::checkInter_rd0_4(Mode& interMode, const CUGeom& cuGeom, PartSize partSize, uint32_t refMask[2])
 {
     interMode.initCosts();
@@ -3142,6 +3550,32 @@ void Analysis::checkInter_rd5_6(Mode& interMode, const CUGeom& cuGeom, PartSize
     }
 }
 
+void Analysis::checkIntraBC_rd5_6(Mode& intraBCMode, const CUGeom& cuGeom, PartSize ePartSize, bool testOnlyPred, bool bUse1DSearchFor8x8, MV lastIntraBCMv[2], MV* iMVCandList)
+{
+    intraBCMode.initCosts();
+    intraBCMode.cu.setPartSizeSubParts(ePartSize);
+    intraBCMode.cu.setPredModeSubParts(MODE_INTER);
+    intraBCMode.cu.setLumaIntraDirSubParts(DC_IDX, 0, cuGeom.depth);
+    intraBCMode.cu.setChromIntraDirSubParts(DC_IDX, 0, cuGeom.depth);
+    for (int i = 0; i < 2; i++)
+        intraBCMode.cu.m_lastIntraBCMv[i] = lastIntraBCMv[i];
+
+    bool bValid = predIntraBCSearch(intraBCMode, cuGeom, m_csp != X265_CSP_I400 && m_frame->m_fencPic->m_picCsp != X265_CSP_I400, ePartSize, testOnlyPred, bUse1DSearchFor8x8);
+    if (bValid)
+        encodeResAndCalcRdInterCU(intraBCMode, cuGeom);
+    else
+        intraBCMode.rdCost = UINT64_MAX;
+
+    if (bValid && (intraBCMode.cu.m_log2CUSize[0] <= 4) && (intraBCMode.cu.m_partSize[0] == SIZE_2NxN || intraBCMode.cu.m_partSize[0] == SIZE_Nx2N))
+    {
+        int dummyWidth, dummyHeight;
+        uint32_t partAddr = 0;
+        intraBCMode.cu.getPartIndexAndSize(1, partAddr, dummyWidth, dummyHeight);
+        iMVCandList[0] = intraBCMode.cu.m_mv[0][0];
+        iMVCandList[1] = intraBCMode.cu.m_mv[0][partAddr];
+    }
+}
+
 void Analysis::checkBidir2Nx2N(Mode& inter2Nx2N, Mode& bidir2Nx2N, const CUGeom& cuGeom)
 {
     CUData& cu = bidir2Nx2N.cu;
@@ -3298,7 +3732,7 @@ void Analysis::encodeResidue(const CUData& ctu, const CUGeom& cuGeom)
 
     cu.copyFromPic(ctu, cuGeom, m_csp);
 
-    PicYuv& reconPic = *m_frame->m_reconPic;
+    PicYuv& reconPic = *m_frame->m_reconPic[0];
 
     Yuv& fencYuv = m_modeDepth[cuGeom.depth].fencYuv;
     if (cuGeom.depth)
@@ -3791,3 +4225,37 @@ int Analysis::findSameContentRefCount(const CUData& parentCTU, const CUGeom& cuG
     }
     return sameContentRef;
 }
+
+int Analysis::CalculateMinimumHVLumaActivity(const CUGeom & cuGeom, const uint32_t uiAbsPartIdx, Yuv ppcOrigYuv)
+{
+    Yuv * pOrgYuv = &ppcOrigYuv;
+    const int      stride = pOrgYuv->m_size;
+    const int      width = 1 << cuGeom.log2CUSize;
+    const int      height = 1 << cuGeom.log2CUSize;
+
+    // Get activity
+    int hAct = 0;
+    const pixel * pY = pOrgYuv->getLumaAddr(uiAbsPartIdx);
+    for (int y = 0; y < height; y++)
+    {
+        for (int x = 1; x < width; x++)
+        {
+            hAct += abs(pY[x] - pY[x - 1]);
+        }
+        pY += stride;
+    }
+
+    int vAct = 0;
+    pY = pOrgYuv->getLumaAddr(0) + stride;
+    for (int y = 1; y < height; y++)
+    {
+        for (int x = 0; x < width; x++)
+        {
+            vAct += abs(pY[x] - pY[x - stride]);
+        }
+        pY += stride;
+    }
+
+    return min(hAct, vAct);
+}
+
diff --git a/source/encoder/analysis.h b/source/encoder/analysis.h
index 3bcb56bc3..6f2e4d931 100644
--- a/source/encoder/analysis.h
+++ b/source/encoder/analysis.h
@@ -75,6 +75,12 @@ public:
         PRED_nRx2N,
         PRED_INTRA_NxN, /* 4x4 intra PU blocks for 8x8 CU */
         PRED_LOSSLESS,  /* lossless encode of best mode */
+        PRED_IBC_2Nx2N,
+        PRED_IBC_Nx2N,
+        PRED_IBC_2NxN,
+        PRED_MIXED_IBC_NX2N,
+        PRED_MIXED_IBC_2NXN,
+        PRED_MERGE_IBC,
         MAX_PRED_TYPES
     };
 
@@ -120,6 +126,8 @@ public:
 
     Mode& compressCTU(CUData& ctu, Frame& frame, const CUGeom& cuGeom, const Entropy& initialContext);
     int32_t loadTUDepth(CUGeom cuGeom, CUData parentCTU);
+
+    static int CalculateMinimumHVLumaActivity(const CUGeom& cuGeom, const uint32_t uiAbsPartIdx, Yuv ppcOrigYuv);
 protected:
     /* Analysis data for save/load mode, writes/reads data based on absPartIdx */
     x265_analysis_inter_data*  m_reuseInterDataCTU;
@@ -181,6 +189,9 @@ protected:
 
     void checkBidir2Nx2N(Mode& inter2Nx2N, Mode& bidir2Nx2N, const CUGeom& cuGeom);
 
+    void checkRDCostIntraBCMerge2Nx2N(Mode& merge, const CUGeom& cuGeom);
+    void checkIntraBC_rd5_6(Mode& intraBCMode, const CUGeom& cuGeom, PartSize ePartSize, bool testOnlyPred, bool bUse1DSearchFor8x8, MV lastIntraBCMv[2], MV* iMVCandList = (0, 0));
+
     /* encode current bestMode losslessly, pick best RD cost */
     void tryLossless(const CUGeom& cuGeom);
 
diff --git a/source/encoder/dpb.cpp b/source/encoder/dpb.cpp
index e488cebb8..d957008eb 100644
--- a/source/encoder/dpb.cpp
+++ b/source/encoder/dpb.cpp
@@ -325,6 +325,48 @@ void DPB::prepareEncode(Frame *newFrame)
         slice->m_colRefIdx = 0;
     }
 
+
+    bool bGPBcheck = false;
+    if (slice->m_sliceType == B_SLICE)
+    {
+        if (slice->m_param->bEnableSCC)
+        {
+            if (slice->m_numRefIdx[0] - 1 == slice->m_numRefIdx[1])
+            {
+                bGPBcheck = true;
+                for (int i = 0; i < slice->m_numRefIdx[1]; i++)
+                {
+                    if (slice->m_refPOCList[1][i] != slice->m_refPOCList[0][i])
+                    {
+                        bGPBcheck = false;
+                        break;
+                    }
+                }
+            }
+        }
+        else if (slice->m_numRefIdx[0] == slice->m_numRefIdx[1])
+        {
+            bGPBcheck = true;
+            int i;
+            for (i = 0; i < slice->m_numRefIdx[1]; i++)
+            {
+                if (slice->m_refPOCList[1][i] != slice->m_refPOCList[0][i])
+                {
+                    bGPBcheck = false;
+                    break;
+                }
+            }
+        }
+    }
+    if (bGPBcheck)
+    {
+        slice->m_bLMvdL1Zero = true;
+    }
+    else
+    {
+        slice->m_bLMvdL1Zero = false;
+    }
+
     // Disable Loopfilter in bound area, because we will do slice-parallelism in future
     slice->m_sLFaseFlag = (newFrame->m_param->maxSlices > 1) ? false : ((SLFASE_CONSTANT & (1 << (pocCurr % 31))) > 0);
 
diff --git a/source/encoder/entropy.cpp b/source/encoder/entropy.cpp
index c79c3ea0d..2ec93ab6a 100644
--- a/source/encoder/entropy.cpp
+++ b/source/encoder/entropy.cpp
@@ -1037,7 +1037,7 @@ void Entropy::codeSliceHeader(const Slice& slice, FrameData& encData, uint32_t s
     }
 
     if (slice.isInterB())
-        WRITE_FLAG(0, "mvd_l1_zero_flag");
+        WRITE_FLAG(slice.m_bLMvdL1Zero, "mvd_l1_zero_flag");
 
     if (slice.m_sps->bTemporalMVPEnabled)
     {
diff --git a/source/encoder/search.cpp b/source/encoder/search.cpp
index 79af0b350..ba220eaa6 100644
--- a/source/encoder/search.cpp
+++ b/source/encoder/search.cpp
@@ -1896,19 +1896,7 @@ uint32_t Search::mergeEstimation(CUData& cu, const CUGeom& cuGeom, const Predict
     MVField  candMvField[MRG_MAX_NUM_CANDS][2];
     uint8_t  candDir[MRG_MAX_NUM_CANDS];
     uint32_t numMergeCand = cu.getInterMergeCandidates(pu.puAbsPartIdx, puIdx, candMvField, candDir);
-
-    if (cu.isBipredRestriction())
-    {
-        /* do not allow bidir merge candidates if PU is smaller than 8x8, drop L1 reference */
-        for (uint32_t mergeCand = 0; mergeCand < numMergeCand; ++mergeCand)
-        {
-            if (candDir[mergeCand] == 3)
-            {
-                candDir[mergeCand] = 1;
-                candMvField[mergeCand][1].refIdx = REF_NOT_VALID;
-            }
-        }
-    }
+    restrictBipredMergeCand(&cu, 0, candMvField, candDir, numMergeCand);
 
     Yuv& tempYuv = m_rqt[cuGeom.depth].tmpPredYuv;
 
@@ -1937,6 +1925,10 @@ uint32_t Search::mergeEstimation(CUData& cu, const CUGeom& cuGeom, const Predict
                 continue;
         }
 
+        if ((candDir[mergeCand] == 1 || candDir[mergeCand] == 3) && (m_slice->m_refPOCList[0][candMvField[mergeCand][0].refIdx] == m_slice->m_poc))
+        {
+            continue;
+        }
         cu.m_mv[0][pu.puAbsPartIdx] = candMvField[mergeCand][0].mv;
         cu.m_refIdx[0][pu.puAbsPartIdx] = (int8_t)candMvField[mergeCand][0].refIdx;
         cu.m_mv[1][pu.puAbsPartIdx] = candMvField[mergeCand][1].mv;
@@ -2664,6 +2656,1784 @@ void Search::predInterSearch(Mode& interMode, const CUGeom& cuGeom, bool bChroma
     interMode.sa8dBits += totalmebits;
 }
 
+uint32_t Search::getSAD(pixel* ref, int refStride, const pixel* curr, int currStride, int width, int height)
+{
+    uint32_t dist = 0;
+
+    for (int i = 0; i < height; i++)
+    {
+        for (int j = 0; j < width; j++)
+        {
+            dist += abs(ref[j] - curr[j]);
+        }
+        ref += refStride;
+        curr += currStride;
+    }
+    return dist;
+}
+
+int Search::intraBCSearchMVChromaRefine(Mode& intraBCMode,
+    const CUGeom& cuGeom,
+    int         roiWidth,
+    int         roiHeight,
+    int         cuPelX,
+    int         cuPelY,
+    uint32_t* sadBestCand,
+    MV* MVCand,
+    uint32_t    partOffset,
+    int         puIdx
+)
+{
+    int bestCandIdx = 0;
+    uint32_t  sadBest = UINT_MAX;
+    uint32_t  tempSad;
+
+    pixel* ref;
+    const pixel* picOrg;
+    int refStride, orgStride;
+    int width, height;
+
+    int picWidth = m_slice->m_sps->picWidthInLumaSamples;
+    int picHeight = m_slice->m_sps->picHeightInLumaSamples;
+
+    CUData& cu = intraBCMode.cu;
+    Yuv& tmpPredYuv = intraBCMode.predYuv;
+    PredictionUnit pu(cu, cuGeom, puIdx);
+
+    for (int cand = 0; cand < CHROMA_REFINEMENT_CANDIDATES; cand++)
+    {
+        if ((!MVCand[cand].x) && (!MVCand[cand].y))
+        {
+            continue;
+        }
+
+        if (((int)(cuPelY + MVCand[cand].y + roiHeight) >= picHeight) || ((cuPelY + MVCand[cand].y) < 0))
+        {
+            continue;
+        }
+
+        if (((int)(cuPelX + MVCand[cand].x + roiWidth) >= picWidth) || ((cuPelX + MVCand[cand].x) < 0))
+        {
+            continue;
+        }
+
+        tempSad = sadBestCand[cand];
+        int bitDepths = m_param->sourceBitDepth;
+        MV mvQuaterPixl = MVCand[cand];
+        mvQuaterPixl <<= 2;
+        cu.setPUMv(0, mvQuaterPixl, pu.puAbsPartIdx, puIdx);
+        cu.setPURefIdx(0, m_slice->m_numRefIdx[0] - 1, pu.puAbsPartIdx, puIdx);
+        cu.setPUMv(1, MV(), pu.puAbsPartIdx, puIdx);
+        cu.setPURefIdx(1, -1, pu.puAbsPartIdx, puIdx);
+        cu.setPUInterDir(1, pu.puAbsPartIdx, puIdx);
+
+        motionCompensation(cu, pu, tmpPredYuv, 1, 1);
+
+        for (uint32_t ch = TEXT_CHROMA_U; ch < MAX_NUM_COMPONENT; ch++)
+        {
+            ref = m_slice->m_refFrameList[0][m_slice->m_numRefIdx[0] - 1]->m_reconPic->getChromaAddr(ch, cu.m_cuAddr, cu.m_absIdxInCTU + partOffset);
+
+            picOrg = intraBCMode.fencYuv->getChromaAddr(ch, partOffset);
+            orgStride = intraBCMode.fencYuv->m_csize;
+
+            refStride = m_frame->m_reconPic->m_strideC;
+
+            width = roiWidth >> m_hChromaShift;
+            height = roiHeight >> m_vChromaShift;
+
+            ref = tmpPredYuv.getChromaAddr(ch, partOffset);
+            refStride = tmpPredYuv.m_csize;
+
+            for (int row = 0; row < height; row++)
+            {
+                for (int col = 0; col < width; col++)
+                {
+                    tempSad += ((abs(ref[col] - picOrg[col])) >> (bitDepths - 8));
+                }
+                ref += refStride;
+                picOrg += orgStride;
+            }
+        }
+
+        if (tempSad < sadBest)
+        {
+            sadBest = tempSad;
+            bestCandIdx = cand;
+        }
+    }
+
+    return bestCandIdx;
+}
+
+void Search::updateBVMergeCandLists(int roiWidth, int roiHeight, MV* mvCand)
+{
+    if (roiWidth + roiHeight > 8)
+    {
+        m_numBVs = mergeCandLists(m_BVs, m_numBVs, mvCand, CHROMA_REFINEMENT_CANDIDATES, false);
+
+        if (roiWidth + roiHeight == 32)
+        {
+            m_numBV16s = m_numBVs;
+        }
+    }
+}
+
+void Search::intraBCSearchMVCandUpdate(uint32_t sad, int x, int y, uint32_t* sadBestCand, MV* MVCand)
+{
+    int j = CHROMA_REFINEMENT_CANDIDATES - 1;
+
+    if (sad < sadBestCand[CHROMA_REFINEMENT_CANDIDATES - 1])
+    {
+        for (int t = CHROMA_REFINEMENT_CANDIDATES - 1; t >= 0; t--)
+        {
+            if (sad < sadBestCand[t])
+            {
+                j = t;
+            }
+        }
+
+        for (int k = CHROMA_REFINEMENT_CANDIDATES - 1; k > j; k--)
+        {
+            sadBestCand[k] = sadBestCand[k - 1];
+
+            MVCand[k].set(MVCand[k - 1].x, MVCand[k - 1].y);
+        }
+        sadBestCand[j] = sad;
+        MVCand[j].set(x, y);
+    }
+}
+
+uint32_t Search::mergeCandLists(MV* dst, uint32_t dn, MV* src, uint32_t sn, bool isSrcQuarPel)
+{
+    for (uint32_t cand = 0; cand < sn && dn < SCM_S0067_NUM_CANDIDATES; cand++)
+    {
+        bool found = false;
+        MV TempMv = src[cand];
+        if (!isSrcQuarPel)
+        {
+            TempMv <<= 2;
+        }
+        for (int j = 0; j < dn; j++)
+        {
+            if (TempMv == dst[j])
+            {
+                found = true;
+                break;
+            }
+        }
+
+        if (!found)
+        {
+            dst[dn] = TempMv;
+            dn++;
+        }
+    }
+    return dn;
+}
+
+void Search::restrictBipredMergeCand(CUData* cu, uint32_t puIdx, MVField(*mvFieldNeighbours)[2], uint8_t* interDirNeighbours, int numValidMergeCand)
+{
+    {
+        for (uint32_t mergeCand = 0; mergeCand < numValidMergeCand; ++mergeCand)
+        {
+            if (interDirNeighbours[mergeCand] == 3)
+            {
+                bool b8x8BiPredRestricted = cu->is8x8BipredRestriction(
+                    mvFieldNeighbours[mergeCand][0].mv,
+                    mvFieldNeighbours[mergeCand][1].mv,
+                    mvFieldNeighbours[mergeCand][0].refIdx,
+                    mvFieldNeighbours[mergeCand][1].refIdx);
+
+                int width = 0;
+                int height = 0;
+                uint32_t partAddr;
+
+                cu->getPartIndexAndSize(puIdx, partAddr, width, height);
+                if (b8x8BiPredRestricted)
+                {
+                    if (width <= 8 && height <= 8)
+                    {
+                        interDirNeighbours[mergeCand] = 1;
+                        mvFieldNeighbours[mergeCand][1].refIdx = REF_NOT_VALID;
+                    }
+                }
+                else if (cu->isBipredRestriction())
+                {
+                    interDirNeighbours[mergeCand] = 1;
+                    mvFieldNeighbours[mergeCand][1].refIdx = REF_NOT_VALID;
+                }
+            }
+        }
+    }
+}
+
+bool Search::isBlockVectorValid(int xPos, int yPos, int width, int height, CUData* cu,
+    int xStartInCU, int yStartInCU, int xBv, int yBv, int ctuSize)
+{
+    static const int s_floorLog2[65] =
+    {
+      -1, 0, 1, 1, 2, 2, 2, 2, 3, 3,
+       3, 3, 3, 3, 3, 3, 4, 4, 4, 4,
+       4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
+       4, 4, 5, 5, 5, 5, 5, 5, 5, 5,
+       5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
+       5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
+       5, 5, 5, 5, 6
+    };
+
+    int ctuSizeLog2 = s_floorLog2[ctuSize];
+    int interpolationSamplesX = (cu->m_chromaFormat == X265_CSP_I422 || cu->m_chromaFormat == X265_CSP_I420) ? ((xBv & 0x1) << 1) : 0;
+    int interpolationSamplesY = (cu->m_chromaFormat == X265_CSP_I420) ? ((yBv & 0x1) << 1) : 0;
+    int refRightX = xPos + xBv + width - 1 + interpolationSamplesX;
+    int refBottomY = yPos + yBv + height - 1 + interpolationSamplesY;
+    int picWidth = m_slice->m_sps->picWidthInLumaSamples;
+    int picHeight = m_slice->m_sps->picHeightInLumaSamples;
+
+    if ((xPos + xBv - interpolationSamplesX) < 0)
+        return false;
+    if (refRightX >= picWidth)
+        return false;
+    if ((yPos + yBv - interpolationSamplesY) < 0)
+        return false;
+    if (refBottomY >= picHeight)
+        return false;
+
+    if ((xBv + width + interpolationSamplesX) > 0 && (yBv + height + interpolationSamplesY) > 0)
+        return false;
+
+    if (refBottomY >> ctuSizeLog2 < yPos >> ctuSizeLog2)
+    {
+        int refCuX = refRightX / ctuSize;
+        int refCuY = refBottomY / ctuSize;
+        int cuPelX = xPos / ctuSize;
+        int cuPelY = yPos / ctuSize;
+
+        if (((int)(refCuX - cuPelX) > (int)((cuPelY - refCuY))))
+            return false;
+        else
+            return true;
+    }
+
+    if (refBottomY >> ctuSizeLog2 > yPos >> ctuSizeLog2)
+    {
+        return false;
+    }
+
+    // in the same CTU line
+    if (refRightX >> ctuSizeLog2 < xPos >> ctuSizeLog2)
+        return true;
+    if (refRightX >> ctuSizeLog2 > xPos >> ctuSizeLog2)
+        return false;
+
+    // same CTU
+    int mask = 1 << ctuSizeLog2;
+    mask -= 1;
+    int rasterCurr = ((((yPos & mask) - yStartInCU) >> 2) << (ctuSizeLog2 - 2)) + (((xPos & mask) - xStartInCU) >> 2);
+    int rasterRef = (((refBottomY & mask) >> 2) << (ctuSizeLog2 - 2)) + ((refRightX & mask) >> 2);
+
+    if (g_rasterToZscan[rasterRef] >= g_rasterToZscan[rasterCurr])
+        return false;
+    return true;
+}
+
+bool Search::isValidIntraBCSearchArea(CUData* cu, int predX, int predY, int roiWidth, int roiHeight, int partOffset)
+{
+    const int  cuPelX = cu->m_cuPelX + g_zscanToPelX[partOffset];
+    const int  cuPelY = cu->m_cuPelY + g_zscanToPelY[partOffset];
+
+    if (!isBlockVectorValid(cuPelX, cuPelY, roiWidth, roiHeight, cu, g_zscanToPelX[partOffset], g_zscanToPelY[partOffset], predX, predY, m_param->maxCUSize))
+    {
+        return false;
+    }
+}
+
+void Search::intraPatternSearch(Mode& intraBCMode, const CUGeom& cuGeom, int puIdx, uint32_t partAddr, pixel* refY, int refStride, MV* searchRangeLT, MV* searchRangeRB,
+    MV& mv, uint32_t& cost, int roiWidth, int roiHeight, bool testOnlyPred, bool bUse1DSearchFor8x8)
+{
+    const int   srchRngHorLeft = searchRangeLT->x;
+    const int   srchRngHorRight = searchRangeRB->x;
+    const int   srchRngVerTop = searchRangeLT->y;
+    const int   srchRngVerBottom = searchRangeRB->y;
+
+    CUData& cu = intraBCMode.cu;
+    const uint32_t  lcuWidth = m_param->maxCUSize;
+    const uint32_t  lcuHeight = m_param->maxCUSize;
+    const int       puPelOffsetX = g_zscanToPelX[partAddr];
+    const int       puPelOffsetY = g_zscanToPelY[partAddr];
+    const int       cuPelX = cu.m_cuPelX + puPelOffsetX;  // Point to the location of PU
+    const int       cuPelY = cu.m_cuPelY + puPelOffsetY;
+
+    uint32_t  sad = 0;
+    uint32_t  sadBest = UINT_MAX;
+    int         bestX = 0;
+    int         bestY = 0;
+    pixel* refSrch;
+    pixel* origPic;
+
+    int         bestCandIdx = 0;
+    uint32_t    partOffset = 0;
+    MV          MVCand[CHROMA_REFINEMENT_CANDIDATES];
+    uint32_t    sadBestCand[CHROMA_REFINEMENT_CANDIDATES];
+
+    partOffset = partAddr;
+    PredictionUnit pu(cu, cuGeom, puIdx);
+    for (int cand = 0; cand < CHROMA_REFINEMENT_CANDIDATES; cand++)
+    {
+        sadBestCand[cand] = UINT_MAX;
+        MVCand[cand].set(0, 0);
+    }
+
+    const int         relCUPelX = cuPelX % lcuWidth;
+    const int         relCUPelY = cuPelY % lcuHeight;
+    const int chromaROIWidthInPixels = roiWidth;
+    const int chromaROIHeightInPixels = roiHeight;
+    bool fastsearch = (m_param->bEnableSCC == 1) ? true : false;
+
+    if (fastsearch)
+    {
+        uint32_t tempSadBest = 0;
+        int srLeft = srchRngHorLeft, srRight = srchRngHorRight, srTop = srchRngVerTop, srBottom = srchRngVerBottom;
+        const uint32_t picWidth = m_slice->m_sps->picWidthInLumaSamples;
+        const uint32_t picHeight = m_slice->m_sps->picHeightInLumaSamples;
+
+        if (1)//full frame search
+        {
+            srLeft = -1 * cuPelX;
+            srTop = -1 * cuPelY;
+
+            srRight = picWidth - cuPelX - roiWidth;
+            srBottom = lcuHeight - cuPelY % lcuHeight - roiHeight;
+
+            if (cuPelX + srRight + roiWidth > picWidth)
+            {
+                srRight = picWidth % lcuWidth - cuPelX % lcuWidth - roiWidth;
+            }
+            if (cuPelY + srBottom + roiHeight > picHeight)
+            {
+                srBottom = picHeight % lcuHeight - cuPelY % lcuHeight - roiHeight;
+            }
+        }
+
+        if (roiWidth > 8 || roiHeight > 8)
+            m_numBVs = 0;
+        else if (roiWidth + roiHeight == 16)
+            m_numBVs = m_numBV16s;
+        if (testOnlyPred)
+            m_numBVs = 0;
+
+        MV  mvPredEncOnly[16];
+        int nbPreds = 0;
+
+        cu.getIntraBCMVPsEncOnly(partAddr, mvPredEncOnly, nbPreds, puIdx);
+        m_numBVs = mergeCandLists(m_BVs, m_numBVs, mvPredEncOnly, nbPreds, true);
+
+        for (uint32_t cand = 0; cand < m_numBVs; cand++)
+        {
+            int xPred = m_BVs[cand].x >> 2;
+            int yPred = m_BVs[cand].y >> 2;
+            if (!(xPred == 0 && yPred == 0) && !((yPred < srTop) || (yPred > srBottom)) && !((xPred < srLeft) || (xPred > srRight)))
+            {
+                int tempY = yPred + relCUPelY + roiHeight - 1;
+                int tempX = xPred + relCUPelX + roiWidth - 1;
+                bool validCand = isValidIntraBCSearchArea(&cu, xPred, yPred, chromaROIWidthInPixels, chromaROIHeightInPixels, partOffset);
+
+                if ((tempX >= (int)lcuWidth) && (tempY >= 0) && 1)
+                    validCand = false;
+
+                if ((tempX >= 0) && (tempY >= 0))
+                {
+                    int tempRasterIdx = (tempY / 4) * cu.s_numPartInCUSize + (tempX / 4);
+                    int tempZscanIdx = g_rasterToZscan[tempRasterIdx];
+                    if (tempZscanIdx >= cu.m_absIdxInCTU)
+                    {
+                        validCand = false;
+                    }
+                }
+
+                if (validCand)
+                {
+                    sad = m_me.mvcost(m_BVs[cand]);
+
+                    refSrch = refY + yPred * refStride + xPred;
+
+                    const pixel* curr = intraBCMode.fencYuv->getLumaAddr(partAddr);
+                    intptr_t currStride = intraBCMode.fencYuv->m_size;
+
+                    sad += m_me.bufSAD(refSrch, refStride);
+                    if (sad > sadBestCand[CHROMA_REFINEMENT_CANDIDATES - 1])
+                    {
+                        continue;
+                    }
+
+                    intraBCSearchMVCandUpdate(sad, xPred, yPred, sadBestCand, MVCand);
+                }
+            }
+        }
+        bestX = MVCand[0].x;
+        bestY = MVCand[0].y;
+        mv.set(bestX, bestY);
+        sadBest = sadBestCand[0];
+
+        if (testOnlyPred)
+        {
+            cost = sadBest;
+            return;
+        }
+
+        const int boundY = (0 - roiHeight - puPelOffsetY);
+        int lowY = ((cu.m_partSize[partAddr] == SCM_S0067_IBC_FULL_1D_SEARCH_FOR_PU) && 1)
+            ? -cuPelY : max(srchRngVerTop, 0 - cuPelY);
+        for (int y = boundY; y >= lowY; y--)
+        {
+            if (!isValidIntraBCSearchArea(&cu, 0, y, chromaROIWidthInPixels, chromaROIHeightInPixels, partOffset))
+            {
+                continue;
+            }
+
+            sad = m_me.mvcost(MV(0, y));
+
+            refSrch = refY + y * refStride;
+
+            const pixel* curr = intraBCMode.fencYuv->getLumaAddr(partAddr);
+            intptr_t currStride = intraBCMode.fencYuv->m_size;
+            sad += m_me.bufSAD(refSrch, refStride);
+            if (sad > sadBestCand[CHROMA_REFINEMENT_CANDIDATES - 1])
+            {
+                continue;
+            }
+
+            intraBCSearchMVCandUpdate(sad, 0, y, sadBestCand, MVCand);
+            tempSadBest = sadBestCand[0];
+            if (sadBestCand[0] <= 3)
+            {
+                bestX = MVCand[0].x;
+                bestY = MVCand[0].y;
+                sadBest = sadBestCand[0];
+                mv.set(bestX, bestY);
+                cost = sadBest;
+
+                updateBVMergeCandLists(roiWidth, roiHeight, MVCand);
+                return;
+            }
+        }
+
+        const int boundX = ((cu.m_partSize[partAddr] == SCM_S0067_IBC_FULL_1D_SEARCH_FOR_PU) && 1)
+            ? -cuPelX : max(srchRngHorLeft, -cuPelX);
+        for (int x = 0 - roiWidth - puPelOffsetX; x >= boundX; --x)
+        {
+            if (!isValidIntraBCSearchArea(&cu, x, 0, chromaROIWidthInPixels, chromaROIHeightInPixels, partOffset))
+            {
+                continue;
+            }
+
+            sad = m_me.mvcost(MV(x, 0));
+
+            refSrch = refY + x;
+
+            const pixel* curr = intraBCMode.fencYuv->getLumaAddr(partAddr);
+            intptr_t currStride = intraBCMode.fencYuv->m_size;
+            sad += m_me.bufSAD(refSrch, refStride);
+
+            if (sad > sadBestCand[CHROMA_REFINEMENT_CANDIDATES - 1])
+            {
+                continue;
+            }
+
+            intraBCSearchMVCandUpdate(sad, x, 0, sadBestCand, MVCand);
+            tempSadBest = sadBestCand[0];
+            if (sadBestCand[0] <= 3)
+            {
+                bestX = MVCand[0].x;
+                bestY = MVCand[0].y;
+                sadBest = sadBestCand[0];
+                mv.set(bestX, bestY);
+                cost = sadBest;
+
+                updateBVMergeCandLists(roiWidth, roiHeight, MVCand);
+                return;
+            }
+        }
+
+        bestX = MVCand[0].x;
+        bestY = MVCand[0].y;
+        sadBest = sadBestCand[0];
+
+        if ((!bestX && !bestY) || (sadBest - m_me.mvcost(MV(bestX, bestY)) <= 32))
+        {
+            //chroma refine
+            bestCandIdx = intraBCSearchMVChromaRefine(intraBCMode, cuGeom, roiWidth, roiHeight, cuPelX, cuPelY, sadBestCand, MVCand, partOffset, puIdx);
+            bestX = MVCand[bestCandIdx].x;
+            bestY = MVCand[bestCandIdx].y;
+            sadBest = sadBestCand[bestCandIdx];
+            mv.set(bestX, bestY);
+            cost = sadBest;
+
+            updateBVMergeCandLists(roiWidth, roiHeight, MVCand);
+            return;
+        }
+
+        if (cuGeom.depth > 2 && bUse1DSearchFor8x8)
+        {
+            for (int y = max(srchRngVerTop, -cuPelY); y <= srchRngVerBottom; y += 2)
+            {
+                if ((y == 0) || ((int)(cuPelY + y + roiHeight) >= picHeight))
+                {
+                    continue;
+                }
+
+                int tempY = y + relCUPelY + roiHeight - 1;
+
+                for (int x = max(srchRngHorLeft, -cuPelX); x <= srchRngHorRight; x++)
+                {
+                    if ((x == 0) || ((int)(cuPelX + x + roiWidth) >= picWidth))
+                    {
+                        continue;
+                    }
+
+                    int tempX = x + relCUPelX + roiWidth - 1;
+
+                    if ((tempX >= 0) && (tempY >= 0))
+                    {
+                        int iTempRasterIdx = (tempY / 4) * cu.s_numPartInCUSize + (tempX / 4);
+                        int iTempZscanIdx = g_rasterToZscan[iTempRasterIdx];
+                        if (iTempZscanIdx >= cu.m_absIdxInCTU)
+                        {
+                            continue;
+                        }
+                    }
+
+                    if (!isValidIntraBCSearchArea(&cu, x, y, chromaROIWidthInPixels, chromaROIHeightInPixels, partOffset))
+                    {
+                        continue;
+                    }
+
+                    sad = m_me.mvcost(MV(x, y));
+
+                    refSrch = refY + y * refStride + x;
+
+                    const pixel* curr = intraBCMode.fencYuv->getLumaAddr(partAddr);
+                    intptr_t currStride = intraBCMode.fencYuv->m_size;
+                    sad += m_me.bufSAD(refSrch, refStride);
+
+                    intraBCSearchMVCandUpdate(sad, x, y, sadBestCand, MVCand);
+                }
+            }
+
+            bestX = MVCand[0].x;
+            bestY = MVCand[0].y;
+            sadBest = sadBestCand[0];
+            if (sadBest - m_me.mvcost(MV(bestX, bestY)) <= 16)
+            {
+                //chroma refine
+                bestCandIdx = intraBCSearchMVChromaRefine(intraBCMode, cuGeom, roiWidth, roiHeight, cuPelX, cuPelY, sadBestCand, MVCand, partOffset, puIdx);
+                bestX = MVCand[bestCandIdx].x;
+                bestY = MVCand[bestCandIdx].y;
+                sadBest = sadBestCand[bestCandIdx];
+                mv.set(bestX, bestY);
+                cost = sadBest;
+
+                updateBVMergeCandLists(roiWidth, roiHeight, MVCand);
+                return;
+            }
+
+            for (int y = (max(srchRngVerTop, -cuPelY) + 1); y <= srchRngVerBottom; y += 2)
+            {
+                if ((y == 0) || ((int)(cuPelY + y + roiHeight) >= picHeight))
+                {
+                    continue;
+                }
+
+                int tempY = y + relCUPelY + roiHeight - 1;
+
+                for (int x = max(srchRngHorLeft, -cuPelX); x <= srchRngHorRight; x += 2)
+                {
+                    if ((x == 0) || ((int)(cuPelX + x + roiWidth) >= picWidth))
+                    {
+                        continue;
+                    }
+
+                    int tempX = x + relCUPelX + roiWidth - 1;
+
+                    if ((tempX >= 0) && (tempY >= 0))
+                    {
+                        int tempRasterIdx = (tempY / 4) * cu.s_numPartInCUSize + (tempX / 4);
+                        int tempZscanIdx = g_rasterToZscan[tempRasterIdx];
+                        if (tempZscanIdx >= cu.m_absIdxInCTU)
+                        {
+                            continue;
+                        }
+                    }
+
+                    if (!isValidIntraBCSearchArea(&cu, x, y, chromaROIWidthInPixels, chromaROIHeightInPixels, partOffset))
+                    {
+                        continue;
+                    }
+
+                    sad = m_me.mvcost(MV(x, y));
+
+                    refSrch = refY + y * refStride + x;
+
+                    const pixel* curr = intraBCMode.fencYuv->getLumaAddr(partAddr);
+                    intptr_t currStride = intraBCMode.fencYuv->m_size;
+                    sad += m_me.bufSAD(refSrch, refStride);
+
+                    if (sad > sadBestCand[CHROMA_REFINEMENT_CANDIDATES - 1])
+                    {
+                        continue;
+                    }
+
+                    intraBCSearchMVCandUpdate(sad, x, y, sadBestCand, MVCand);
+                    if (sadBestCand[0] <= 5)
+                    {
+                        //chroma refine & return
+                        bestCandIdx = intraBCSearchMVChromaRefine(intraBCMode, cuGeom, roiWidth, roiHeight, cuPelX, cuPelY, sadBestCand, MVCand, partOffset, puIdx);
+                        bestX = MVCand[bestCandIdx].x;
+                        bestY = MVCand[bestCandIdx].y;
+                        sadBest = sadBestCand[bestCandIdx];
+                        mv.set(bestX, bestY);
+                        cost = sadBest;
+
+                        updateBVMergeCandLists(roiWidth, roiHeight, MVCand);
+                        return;
+                    }
+                }
+            }
+
+            bestX = MVCand[0].x;
+            bestY = MVCand[0].y;
+            sadBest = sadBestCand[0];
+
+            if ((sadBest >= tempSadBest) || ((sadBest - m_me.mvcost(MV(bestX, bestY))) <= 32))
+            {
+                //chroma refine
+                bestCandIdx = intraBCSearchMVChromaRefine(intraBCMode, cuGeom, roiWidth, roiHeight, cuPelX, cuPelY, sadBestCand, MVCand, partOffset, puIdx);
+                bestX = MVCand[bestCandIdx].x;
+                bestY = MVCand[bestCandIdx].y;
+                sadBest = sadBestCand[bestCandIdx];
+                mv.set(bestX, bestY);
+                cost = sadBest;
+
+                updateBVMergeCandLists(roiWidth, roiHeight, MVCand);
+                return;
+            }
+
+            tempSadBest = sadBestCand[0];
+
+
+            for (int y = (max(srchRngVerTop, -cuPelY) + 1); y <= srchRngVerBottom; y += 2)
+            {
+                if ((y == 0) || ((int)(cuPelY + y + roiHeight) >= picHeight))
+                {
+                    continue;
+                }
+
+                int tempY = y + relCUPelY + roiHeight - 1;
+
+                for (int x = (max(srchRngHorLeft, -cuPelX) + 1); x <= srchRngHorRight; x += 2)
+                {
+
+                    if ((x == 0) || ((int)(cuPelX + x + roiWidth) >= picWidth))
+                    {
+                        continue;
+                    }
+
+                    int tempX = x + relCUPelX + roiWidth - 1;
+
+                    if ((tempX >= 0) && (tempY >= 0))
+                    {
+                        int tempRasterIdx = (tempY / 4) * cu.s_numPartInCUSize + (tempX / 4);
+                        int tempZscanIdx = g_rasterToZscan[tempRasterIdx];
+                        if (tempZscanIdx >= cu.m_absIdxInCTU)
+                        {
+                            continue;
+                        }
+                    }
+
+                    if (!isValidIntraBCSearchArea(&cu, x, y, chromaROIWidthInPixels, chromaROIHeightInPixels, partOffset))
+                    {
+                        continue;
+                    }
+
+                    sad = m_me.mvcost(MV(x, y));
+
+                    refSrch = refY + y * refStride + x;
+
+                    const pixel* curr = intraBCMode.fencYuv->getLumaAddr(partAddr);
+                    intptr_t currStride = intraBCMode.fencYuv->m_size;
+                    sad += m_me.bufSAD(refSrch, refStride);
+                    if (sad > sadBestCand[CHROMA_REFINEMENT_CANDIDATES - 1])
+                    {
+                        continue;
+                    }
+
+                    intraBCSearchMVCandUpdate(sad, x, y, sadBestCand, MVCand);
+                    if (sadBestCand[0] <= 5)
+                    {
+                        //chroma refine & return
+                        bestCandIdx = intraBCSearchMVChromaRefine(intraBCMode, cuGeom, roiWidth, roiHeight, cuPelX, cuPelY, sadBestCand, MVCand, partOffset, puIdx);
+                        bestX = MVCand[bestCandIdx].x;
+                        bestY = MVCand[bestCandIdx].y;
+                        sadBest = sadBestCand[bestCandIdx];
+                        mv.set(bestX, bestY);
+                        cost = sadBest;
+
+                        updateBVMergeCandLists(roiWidth, roiHeight, MVCand);
+                        return;
+                    }
+                }
+            }
+        }
+    }
+    else //full search
+    {
+        refY += (srchRngVerBottom * refStride);
+        int picWidth = m_slice->m_sps->picWidthInLumaSamples;
+        int picHeight = m_slice->m_sps->picHeightInLumaSamples;
+
+        for (int y = srchRngVerBottom; y >= srchRngVerTop; y--)
+        {
+            if (((int)(cuPelY + y) < 0) || ((int)(cuPelY + y + roiHeight) >= picHeight))
+            {
+                refY -= refStride;
+                continue;
+            }
+
+            for (int x = srchRngHorLeft; x <= srchRngHorRight; x++)
+            {
+
+                if (((int)(cuPelX + x) < 0) || ((int)(cuPelX + x + roiWidth) >= picWidth))
+                {
+                    continue;
+                }
+
+                int tempX = x + relCUPelX + roiWidth - 1;
+                int tempY = y + relCUPelY + roiHeight - 1;
+                if ((tempX >= 0) && (tempY >= 0))
+                {
+                    int iTempRasterIdx = (tempY / 4) * cu.s_numPartInCUSize + (tempX / 4);
+                    int iTempZscanIdx = g_rasterToZscan[iTempRasterIdx];
+                    if (iTempZscanIdx >= cu.m_absIdxInCTU)
+                    {
+                        continue;
+                    }
+                }
+
+                if (!isValidIntraBCSearchArea(&cu, x, y, chromaROIWidthInPixels, chromaROIHeightInPixels, partOffset))
+                {
+                    continue;
+                }
+
+                refSrch = refY + x;
+
+                sad = m_me.bufSAD(refSrch, refStride);
+                sad += m_me.mvcost(MV(x, y));
+                if (sad < sadBest)
+                {
+                    sadBest = sad;
+                    bestX = x;
+                    bestY = y;
+                }
+                intraBCSearchMVCandUpdate(sad, x, y, sadBestCand, MVCand);
+            }
+
+            refY -= refStride;
+        }
+    }
+
+    bestCandIdx = intraBCSearchMVChromaRefine(intraBCMode, cuGeom, roiWidth, roiHeight, cuPelX, cuPelY, sadBestCand, MVCand, partOffset, puIdx);
+    bestX = MVCand[bestCandIdx].x;
+    bestY = MVCand[bestCandIdx].y;
+    sadBest = sadBestCand[bestCandIdx];
+    mv.set(bestX, bestY);
+    cost = sadBest;
+
+    updateBVMergeCandLists(roiWidth, roiHeight, MVCand);
+
+}
+
+void Search::setIntraSearchRange(Mode& intraBCMode, MV& pred, int puIdx, int roiWidth, int roiHeight, MV& searchRangeLT, MV& searchRangeRB)
+{
+    MV mvPred = pred;
+    CUData& cu = intraBCMode.cu;
+    cu.clipMv(mvPred);
+    int srLeft, srRight, srTop, srBottom;
+    int width, height;
+    uint32_t partAddr;
+
+    cu.getPartIndexAndSize(puIdx, partAddr, width, height);
+
+    const uint32_t lcuWidth = m_param->maxCUSize;
+    const uint32_t lcuHeight = m_param->maxCUSize;
+    const uint32_t cuPelX = cu.m_cuPelX + g_zscanToPelX[partAddr];
+    const uint32_t cuPelY = cu.m_cuPelY + g_zscanToPelY[partAddr];
+
+    const uint32_t picWidth = m_slice->m_sps->picWidthInLumaSamples;
+    const uint32_t picHeight = m_slice->m_sps->picHeightInLumaSamples;
+
+    if (cu.m_cuDepth[0] == 2 && cu.m_partSize[0] == SIZE_2Nx2N && m_param->bEnableSCC == 2)// full frame search
+    {
+        srLeft = -1 * cuPelX;
+        srTop = -1 * cuPelY;
+
+        srRight = picWidth - cuPelX - roiWidth;
+        srBottom = lcuHeight - cuPelY % lcuHeight - roiHeight;
+    }
+    else
+    {
+        uint32_t maxXsr = cuPelX % lcuWidth;
+        uint32_t maxYsr = cuPelY % lcuHeight;
+
+        if (cu.m_chromaFormat == X265_CSP_I420 || cu.m_chromaFormat == X265_CSP_I422) maxXsr &= ~0x4;
+        if (cu.m_chromaFormat == X265_CSP_I420)                                       maxYsr &= ~0x4;
+
+        srLeft = -maxXsr;
+        srTop = -maxYsr;
+
+        srRight = lcuWidth - cuPelX % lcuWidth - roiWidth;
+        srBottom = lcuHeight - cuPelY % lcuHeight - roiHeight;
+    }
+
+    if (cuPelX + srRight + roiWidth > picWidth)
+    {
+        srRight = picWidth % lcuWidth - cuPelX % lcuWidth - roiWidth;
+    }
+    if (cuPelY + srBottom + roiHeight > picHeight)
+    {
+        srBottom = picHeight % lcuHeight - cuPelY % lcuHeight - roiHeight;
+    }
+
+    searchRangeLT.x = srLeft;
+    searchRangeLT.y = srTop;
+    searchRangeRB.x = srRight;
+    searchRangeRB.y = srBottom;
+
+    cu.clipMv(searchRangeLT);
+    cu.clipMv(searchRangeRB);
+
+}
+
+void Search::intraBlockCopyEstimate(Mode& intraBCMode, const CUGeom& cuGeom, int puIdx, MV* pred, MV& mv, uint32_t& cost, bool testOnlyPred, bool bUse1DSearchFor8x8)
+{
+    uint32_t         partAddr;
+    int              roiWidth;
+    int              roiHeight;
+
+    MV   searchRangeLT;
+    MV   searchRangeRB;
+    MV   mvPred = *pred;
+    const MV predictors = *pred;
+
+    CUData& cu = intraBCMode.cu;
+    const Yuv* fencYuv = intraBCMode.fencYuv;
+    cu.getPartIndexAndSize(puIdx, partAddr, roiWidth, roiHeight);
+
+    /* calculate the location of upper-left corner pixel and size of the current PU */
+    int xP, yP, nPSW, nPSH;
+
+    int cuSize = 1 << cu.m_log2CUSize[0];
+    int partMode = cu.m_partSize[0];
+
+    int tmp = partTable[partMode][puIdx][0];
+    nPSW = ((tmp >> 4) * cuSize) >> 2;
+    nPSH = ((tmp & 0xF) * cuSize) >> 2;
+
+    tmp = partTable[partMode][puIdx][1];
+    xP = ((tmp >> 4) * cuSize) >> 2;
+    yP = ((tmp & 0xF) * cuSize) >> 2;
+
+    assert(nPSW == roiWidth);
+    assert(nPSH == roiHeight);
+
+    int ref = m_slice->m_numRefIdx[0] - 1;
+    pixel* refY = m_slice->m_refFrameList[0][ref]->m_reconPic->getLumaAddr(cu.m_cuAddr, cu.m_absIdxInCTU + partAddr);
+    int  strideY = m_slice->m_refFrameList[0][ref]->m_reconPic->m_stride;
+
+    setIntraSearchRange(intraBCMode, mvPred, puIdx, roiWidth, roiHeight, searchRangeLT, searchRangeRB);
+
+    m_me.setMVP(predictors);
+
+    intraPatternSearch(intraBCMode, cuGeom, puIdx, partAddr, refY, strideY, &searchRangeLT, &searchRangeRB, mv, cost, roiWidth, roiHeight, testOnlyPred, bUse1DSearchFor8x8);
+}
+
+bool Search::predIntraBCSearch(Mode& intraBCMode, const CUGeom& cuGeom, bool bChromaMC, PartSize ePartSize, bool testOnlyPred, bool bUse1DSearchFor8x8)
+{
+    MV zeroMv(0, 0);
+    CUData& cu = intraBCMode.cu;
+    Yuv* predYuv = &intraBCMode.predYuv;
+    Yuv& tmpPredYuv = m_rqt[cuGeom.depth].tmpPredYuv;
+    int  numPart = cu.getNumPartInter(0);
+
+    MV mvc[(MD_ABOVE_LEFT + 1) * 2 + 2];
+
+    if (m_param->bEnableSCC == 1 && cuGeom.depth < 1) // fast search
+        return false;
+
+    uint32_t totalCost = 0;
+    for (int puIdx = 0; puIdx < numPart; puIdx++)
+    {
+        int width, height;
+        uint32_t partAddr = 0;
+        MotionData* bestME = intraBCMode.bestME[puIdx];
+        PredictionUnit pu(cu, cuGeom, puIdx);
+        MV  mv, mvd, mvPred[2];
+        cu.getPartIndexAndSize(puIdx, pu.puAbsPartIdx, width, height);
+        partAddr = pu.puAbsPartIdx;
+        bool bChromaMC = 1;
+        m_me.setSourcePU(*intraBCMode.fencYuv, pu.ctuAddr, pu.cuAbsPartIdx, pu.puAbsPartIdx, pu.width, pu.height, m_param->searchMethod, m_param->subpelRefine, bChromaMC);
+
+        cu.getNeighbourMV(puIdx, pu.puAbsPartIdx, intraBCMode.interNeighbours);
+        int numMvc = cu.getPMV(intraBCMode.interNeighbours, 0, m_slice->m_numRefIdx[0] - 1, intraBCMode.amvpCand[0][m_slice->m_numRefIdx[0] - 1], mvc);
+
+        mvPred[0].set(intraBCMode.amvpCand[0][m_slice->m_numRefIdx[0] - 1][0].x >> 2, intraBCMode.amvpCand[0][m_slice->m_numRefIdx[0] - 1][0].y >> 2);
+        mvPred[1].set(intraBCMode.amvpCand[0][m_slice->m_numRefIdx[0] - 1][1].x >> 2, intraBCMode.amvpCand[0][m_slice->m_numRefIdx[0] - 1][1].y >> 2);
+
+        MVField mvField;
+        uint32_t cost;
+        mv.set(0, 0);
+        intraBlockCopyEstimate(intraBCMode, cuGeom, puIdx, mvPred, mv, cost, testOnlyPred, bUse1DSearchFor8x8);
+
+        bestME->mv.set(mv.x << 2, mv.y << 2);
+        bestME->cost = cost;
+        totalCost += cost;
+        if (mv.x == 0 && mv.y == 0)
+        {
+            if (testOnlyPred)
+            {
+                m_lastCandCost = MAX_UINT;
+            }
+            return false;
+        }
+
+        uint32_t depth = cu.m_cuDepth[0];
+        int bitsAMVPBest, bitsAMVPTemp, bitsMergeTemp;
+        int distAMVPBest, distMergeTemp;
+        int costAMVPBest, costMergeBest, costMergeTemp;
+        bitsAMVPBest = MAX_INT;
+        costAMVPBest = MAX_INT;
+        costMergeBest = MAX_INT;
+        int mvpIdxBest = 0;
+        int mvpIdxTemp;
+        int mrgIdxBest = -1;
+        int mrgIdxTemp = -1;
+        int xCUStart = cu.m_cuPelX;
+        int yCUStart = cu.m_cuPelY;
+        int xStartInCU, yStartInCU;
+        if (ePartSize == SIZE_2Nx2N)
+            xStartInCU = yStartInCU = 0;
+        else if (ePartSize == SIZE_2NxN)
+        {
+            xStartInCU = 0;
+            yStartInCU = (1 << cu.m_log2CUSize[0]) / 2 * puIdx;
+        }
+        else if (ePartSize == SIZE_Nx2N)
+        {
+            xStartInCU = (1 << cu.m_log2CUSize[0]) / 2 * puIdx;
+            yStartInCU = 0;
+        }
+        const pixel* currStart;
+        pixel* curr;
+        pixel* ref;
+        int currStride, refStride;
+        distAMVPBest = 0;
+
+        MV cMvQuaterPixl = mv;
+        cMvQuaterPixl <<= 2;
+        cu.setPUMv(0, cMvQuaterPixl, pu.puAbsPartIdx, puIdx);
+        cu.setPURefIdx(0, (int8_t)m_slice->m_numRefIdx[0] - 1, pu.puAbsPartIdx, puIdx);
+        cu.setPUMv(1, MV(0, 0), pu.puAbsPartIdx, puIdx);
+        cu.setPURefIdx(1, -1, pu.puAbsPartIdx, puIdx);
+        cu.setPUInterDir(1, pu.puAbsPartIdx, puIdx);
+        motionCompensation(cu, pu, tmpPredYuv, 1, 1);
+
+        for (uint32_t ch = TEXT_LUMA; ch < MAX_NUM_COMPONENT; ch++)
+        {
+            int tempHeight, tempWidth;
+            if (ch == 0)
+            {
+                tempHeight = height;
+                tempWidth = width;
+                currStart = intraBCMode.fencYuv->getLumaAddr(partAddr);
+                currStride = intraBCMode.fencYuv->m_size;
+                ref = tmpPredYuv.getLumaAddr(partAddr);
+                refStride = tmpPredYuv.m_size;
+            }
+            else
+            {
+                tempHeight = height >> m_vChromaShift;
+                tempWidth = width >> m_hChromaShift;
+
+                currStart = intraBCMode.fencYuv->getChromaAddr(ch, partAddr);
+                currStride = intraBCMode.fencYuv->m_csize;
+                ref = tmpPredYuv.getChromaAddr(ch, partAddr);
+                refStride = tmpPredYuv.m_csize;
+            }
+            distAMVPBest += getSAD(ref, refStride, currStart, currStride, tempWidth, tempHeight);
+        }
+
+        mvPred[0].set(intraBCMode.amvpCand[0][m_slice->m_numRefIdx[0] - 1][0].x >> 2, intraBCMode.amvpCand[0][m_slice->m_numRefIdx[0] - 1][0].y >> 2);
+        mvPred[1].set(intraBCMode.amvpCand[0][m_slice->m_numRefIdx[0] - 1][1].x >> 2, intraBCMode.amvpCand[0][m_slice->m_numRefIdx[0] - 1][1].y >> 2);
+
+        MV check;
+        for (mvpIdxTemp = 0; mvpIdxTemp < AMVP_NUM_CANDS; mvpIdxTemp++)
+        {
+            if (!(mvPred[mvpIdxTemp].x == check.x >> 2))
+            {
+                m_me.setMVP(mvPred[mvpIdxTemp]);
+                bitsAMVPTemp = m_me.bitcost(mv, mvPred[mvpIdxTemp]);
+                if (bitsAMVPTemp < bitsAMVPBest)
+                {
+                    bitsAMVPBest = bitsAMVPTemp;
+                    mvpIdxBest = mvpIdxTemp;
+                }
+            }
+        }
+
+        bitsAMVPBest++; // for MVP Index bits
+        costAMVPBest = distAMVPBest + m_rdCost.getCost(bitsAMVPBest);
+
+        MVField cMvFieldNeighbours[MRG_MAX_NUM_CANDS][2]; // double length for mv of both lists
+        uint8_t uhInterDirNeighbours[MRG_MAX_NUM_CANDS];
+        int numValidMergeCand = 0;
+
+        for (int i = 0; i < MRG_MAX_NUM_CANDS; i++)
+        {
+            cMvFieldNeighbours[i][0].mv.set(0, 0);
+            cMvFieldNeighbours[i][0].refIdx = REF_NOT_VALID;
+        }
+
+        if (ePartSize != SIZE_2Nx2N)
+        {
+            if (0 && ePartSize != SIZE_2Nx2N && cu.m_cuDepth[0] >= 3)
+            {
+                cu.setPartSizeSubParts(SIZE_2Nx2N);
+                if (puIdx == 0)
+                {
+                    numValidMergeCand = cu.getInterMergeCandidates(0, 0, cMvFieldNeighbours, uhInterDirNeighbours);
+                }
+                cu.setPartSizeSubParts(ePartSize);
+            }
+            else
+            {
+                numValidMergeCand = cu.getInterMergeCandidates(pu.puAbsPartIdx, puIdx, cMvFieldNeighbours, uhInterDirNeighbours);
+            }
+
+            cu.roundMergeCandidates(cMvFieldNeighbours, numValidMergeCand);
+            restrictBipredMergeCand(&cu, puIdx, cMvFieldNeighbours, uhInterDirNeighbours, numValidMergeCand);
+
+            for (mrgIdxTemp = 0; mrgIdxTemp < numValidMergeCand; mrgIdxTemp++)
+            {
+                if (uhInterDirNeighbours[mrgIdxTemp] != 1)
+                {
+                    continue;
+                }
+                if (m_slice->m_refPOCList[0][cMvFieldNeighbours[mrgIdxTemp][0].refIdx] != m_slice->m_poc)
+                {
+                    continue;
+                }
+
+                if (!isBlockVectorValid(xCUStart + xStartInCU, yCUStart + yStartInCU, width, height, &cu,
+                    xStartInCU, yStartInCU, (cMvFieldNeighbours[mrgIdxTemp][0].mv.x >> 2), (cMvFieldNeighbours[mrgIdxTemp][0].mv.y >> 2), m_param->maxCUSize))
+                {
+                    continue;
+                }
+                bitsMergeTemp = mrgIdxTemp == m_param->maxNumMergeCand ? mrgIdxTemp : mrgIdxTemp + 1;
+                bitsMergeTemp = getTUBits(mrgIdxTemp, m_param->maxNumMergeCand);
+
+                distMergeTemp = 0;
+
+                cu.setPUMv(0, cMvFieldNeighbours[mrgIdxTemp][0].mv, pu.puAbsPartIdx, puIdx);
+                cu.setPURefIdx(0, (int8_t)(m_slice->m_numRefIdx[0] - 1), pu.puAbsPartIdx, puIdx);
+                cu.setPUMv(1, MV(0, 0), pu.puAbsPartIdx, puIdx);
+                cu.setPURefIdx(1, -1, pu.puAbsPartIdx, puIdx);
+                cu.setPUInterDir(1, pu.puAbsPartIdx, puIdx);
+                motionCompensation(cu, pu, tmpPredYuv, 1, 1);
+
+                for (int ch = TEXT_LUMA; ch < MAX_NUM_COMPONENT; ch++)
+                {
+                    int tempHeight, tempWidth;
+                    if (ch == 0)
+                    {
+                        tempHeight = height;
+                        tempWidth = width;
+                        currStart = intraBCMode.fencYuv->getLumaAddr(partAddr);
+                        currStride = intraBCMode.fencYuv->m_size;
+                        ref = tmpPredYuv.getLumaAddr(partAddr);
+                        refStride = tmpPredYuv.m_size;
+                    }
+                    else
+                    {
+                        tempHeight = height >> m_vChromaShift;
+                        tempWidth = width >> m_hChromaShift;
+
+                        currStart = intraBCMode.fencYuv->getChromaAddr(ch, partAddr);
+                        currStride = intraBCMode.fencYuv->m_csize;
+                        ref = tmpPredYuv.getChromaAddr(ch, partAddr);
+                        refStride = tmpPredYuv.m_csize;
+                    }
+                    distMergeTemp += getSAD(ref, refStride, currStart, currStride, tempWidth, tempHeight);
+
+                }
+                costMergeTemp = distMergeTemp + m_rdCost.getCost(bitsMergeTemp);
+
+                if (costMergeTemp < costMergeBest)
+                {
+                    costMergeBest = costMergeTemp;
+                    mrgIdxBest = mrgIdxTemp;
+                }
+            }
+        }
+        if (costAMVPBest < costMergeBest)
+        {
+            MV zeroMv(0, 0);
+            MV tempmv((mv.x << 2), (mv.y << 2));
+            MVField mvField[2];
+            mvField[0].mv = tempmv;
+            mvField[0].refIdx = m_slice->m_numRefIdx[0] - 1;   // the current picture is at the last position of list0
+            mvField[1].mv = zeroMv;
+            mvField[1].refIdx = -1;
+
+            cu.m_mergeFlag[pu.puAbsPartIdx] = false;
+            cu.setPUInterDir(1, pu.puAbsPartIdx, puIdx);  // list 0 prediction
+
+            cu.setPUMv(0, mvField[0].mv, pu.puAbsPartIdx, puIdx);
+            cu.setPURefIdx(0, (int8_t)mvField[0].refIdx, pu.puAbsPartIdx, puIdx);
+            cu.setPUMv(1, mvField[1].mv, pu.puAbsPartIdx, puIdx);
+            cu.setPURefIdx(1, (int8_t)mvField[1].refIdx, pu.puAbsPartIdx, puIdx);
+
+            MV mvd;
+            mvd.set(mv.x - (intraBCMode.amvpCand[0][m_slice->m_numRefIdx[0] - 1][mvpIdxBest].x >> 2), mv.y - (intraBCMode.amvpCand[0][m_slice->m_numRefIdx[0] - 1][mvpIdxBest].y >> 2));
+
+            cu.m_mvd[0][pu.puAbsPartIdx] = mvd;
+            cu.m_mvpIdx[0][pu.puAbsPartIdx] = (uint8_t)mvpIdxBest;
+
+            cu.m_mvd[1][pu.puAbsPartIdx] = zeroMv;
+            cu.m_mvpIdx[1][pu.puAbsPartIdx] = REF_NOT_VALID;
+        }
+        else
+        {
+            MV mv(cMvFieldNeighbours[mrgIdxBest][0].mv.x, cMvFieldNeighbours[mrgIdxBest][0].mv.y);
+            MVField mvField[2];
+            mvField[0].mv = mv;
+            mvField[0].refIdx = cu.m_slice->m_numRefIdx[0] - 1;   // the current picture is at the last position of list0
+            mvField[1].mv = zeroMv;
+            mvField[1].refIdx = -1;
+
+            cu.m_mergeFlag[pu.puAbsPartIdx] = true;
+            cu.m_mvpIdx[0][pu.puAbsPartIdx] = (uint8_t)mrgIdxBest; /* merge candidate ID is stored in L0 MVP idx */
+            cu.setPUInterDir(1, pu.puAbsPartIdx, puIdx);  // list 0 prediction
+
+            cu.setPUMv(0, mvField[0].mv, pu.puAbsPartIdx, puIdx);
+            cu.setPURefIdx(0, (int8_t)mvField[0].refIdx, pu.puAbsPartIdx, puIdx);
+            cu.setPUMv(1, mvField[1].mv, pu.puAbsPartIdx, puIdx);
+            cu.setPURefIdx(1, (int8_t)mvField[1].refIdx, pu.puAbsPartIdx, puIdx);
+
+            cu.m_mvd[0][pu.puAbsPartIdx] = zeroMv;
+            cu.m_mvd[1][pu.puAbsPartIdx] = zeroMv;
+        }
+        motionCompensation(cu, pu, *predYuv, 1, 1);
+    }
+
+    PredictionUnit pu(cu, cuGeom, 0);
+    uint32_t abortThreshold = (1 << cu.m_log2CUSize[0]) * (1 << cu.m_log2CUSize[0]) * 2;
+    if (testOnlyPred)
+    {
+        if (numPart == 1 && totalCost > abortThreshold)
+        {
+            m_lastCandCost = MAX_UINT;
+            return false;
+        }
+        m_lastCandCost = totalCost;
+    }
+    else if (totalCost < abortThreshold && 3 * totalCost >> 2 >= m_lastCandCost)
+    {
+        return false;
+    }
+    return true;
+}
+
+bool Search::predMixedIntraBCInterSearch(Mode& intraBCMixedMode, const CUGeom& cuGeom, bool bChromaMC, PartSize ePartSize, bool testOnlyPred, MV* iMvCandList)
+{
+    intraBCMixedMode.initCosts();
+    intraBCMixedMode.cu.setPartSizeSubParts(ePartSize);
+    intraBCMixedMode.cu.setPredModeSubParts(MODE_INTER);
+    CUData& cu = intraBCMixedMode.cu;
+    uint32_t depth = cuGeom.depth;
+    int numComb = 2;
+    int numPart = 2;
+    uint32_t cost[2] = { 0,0 };
+    uint32_t maxCost = UINT32_MAX;
+
+    MVField  cMvFieldNeighbours[MRG_MAX_NUM_CANDS][2]; // double length for mv of both lists
+    uint8_t  uhInterDirNeighbours[MRG_MAX_NUM_CANDS];
+    int      numValidMergeCand[2] = { MRG_MAX_NUM_CANDS, MRG_MAX_NUM_CANDS };
+    int      numPredDir = m_slice->isInterP() ? 1 : 2;
+    MV       cMvZero(0, 0);
+
+    MV  cMvPredCand[2][2];
+    int IBCValidFlag = 0;
+    int bestIBCMvpIdx[2] = { 0, 0 };
+    int bestInterMvpIdx[2] = { 0, 0 };
+    int bestInterDir[2] = { 0, 0 };
+    int bestRefIdx[2] = { 0, 0 };
+    bool isMergeMode[2] = { false, false };
+    bool isIBCMergeMode[2] = { false, false };
+    MVField cMRGMvField[2][2];
+    MVField cMRGMvFieldIBC[2][2];
+
+    // 12 mv candidates including lowresMV
+    MV mvc[(MD_ABOVE_LEFT + 1) * 2 + 2];
+
+    Yuv* predYuv = &intraBCMixedMode.predYuv;
+    Yuv& tmpPredYuv = m_rqt[cuGeom.depth].tmpPredYuv;
+
+    for (int combo = 0; combo < numComb; combo++)
+    {
+        for (int partIdx = 0; partIdx < numPart; ++partIdx)
+        {
+            int dummyWidth, dummyHeight;
+            uint32_t partAddr = 0;
+            PredictionUnit pu(cu, cuGeom, partIdx);
+            cu.getPartIndexAndSize(partIdx, partAddr, dummyWidth, dummyHeight);
+            m_me.setSourcePU(*intraBCMixedMode.fencYuv, pu.ctuAddr, pu.cuAbsPartIdx, pu.puAbsPartIdx, pu.width, pu.height, m_param->searchMethod, m_param->subpelRefine, bChromaMC);
+
+            MV mvPred[2];
+            MV bvPred[2];
+            uint32_t   biPDistTemp = UINT32_MAX;
+            if ((combo == 0 && partIdx == 0) || (combo == 1 && partIdx == 1)) // intraBC
+            {
+                MV cMv = iMvCandList[8 + partIdx];
+                if (cMv.x == 0 && cMv.y == 0)
+                {
+                    cost[combo] = maxCost;
+                    IBCValidFlag++;
+                    break;
+                }
+
+                cu.getNeighbourMV(partIdx, pu.puAbsPartIdx, intraBCMixedMode.interNeighbours);
+                int numMvc = cu.getPMV(intraBCMixedMode.interNeighbours, 0, m_slice->m_numRefIdx[0] - 1, intraBCMixedMode.amvpCand[0][m_slice->m_numRefIdx[0] - 1], mvc);
+
+                bvPred[0] = intraBCMixedMode.amvpCand[0][m_slice->m_numRefIdx[0] - 1][0];
+                bvPred[1] = intraBCMixedMode.amvpCand[0][m_slice->m_numRefIdx[0] - 1][1];
+                bvPred[0] >>= 2;
+                bvPred[1] >>= 2;
+
+                int bitsAMVPBest, bitsAMVPTemp, bitsMergeTemp;
+                int distAMVPBest, distMergeTemp;
+                int costAMVPBest, costMergeBest, costMergeTemp;
+                bitsAMVPBest = MAX_INT;
+                costAMVPBest = MAX_INT;
+                costMergeBest = MAX_INT;
+                int mvpIdxBest = 0;
+                int mvpIdxTemp;
+                int mrgIdxBest = -1;
+                int mrgIdxTemp = -1;
+                int xCUStart = cu.m_cuPelX;
+                int yCUStart = cu.m_cuPelY;
+                int xStartInCU, yStartInCU;
+                if (ePartSize == SIZE_2Nx2N)
+                    xStartInCU = yStartInCU = 0;
+                else if (ePartSize == SIZE_2NxN)
+                {
+                    xStartInCU = 0;
+                    yStartInCU = (1 << cu.m_log2CUSize[0]) / 2 * partIdx;
+                }
+                else if (ePartSize == SIZE_Nx2N)
+                {
+                    xStartInCU = (1 << cu.m_log2CUSize[0]) / 2 * partIdx;
+                    yStartInCU = 0;
+                }
+                const pixel* currStart;
+                int currStride;
+                pixel* pCurr;
+                int refStride;
+                pixel* pRef;
+                distAMVPBest = 0;
+                pixel* ref;
+
+                cu.setPUMv(0, cMv, pu.puAbsPartIdx, partIdx);
+                cu.setPURefIdx(0, (int8_t)m_slice->m_numRefIdx[0] - 1, pu.puAbsPartIdx, partIdx);
+                cu.setPUMv(1, MV(0, 0), pu.puAbsPartIdx, partIdx);
+                cu.setPURefIdx(1, -1, pu.puAbsPartIdx, partIdx);
+                cu.setPUInterDir(1, pu.puAbsPartIdx, partIdx);
+                motionCompensation(cu, pu, tmpPredYuv, 1, 1);
+
+                for (uint32_t ch = TEXT_LUMA; ch < MAX_NUM_COMPONENT; ch++)
+                {
+                    int tempHeight, tempWidth;
+                    if (ch == 0)
+                    {
+                        tempHeight = dummyHeight;
+                        tempWidth = dummyWidth;
+                        currStart = intraBCMixedMode.fencYuv->getLumaAddr(partAddr);
+                        currStride = intraBCMixedMode.fencYuv->m_size;
+                        ref = tmpPredYuv.getLumaAddr(partAddr);
+                        refStride = tmpPredYuv.m_size;
+                    }
+                    else
+                    {
+                        tempHeight = dummyHeight >> m_vChromaShift;
+                        tempWidth = dummyWidth >> m_hChromaShift;
+
+                        currStart = intraBCMixedMode.fencYuv->getChromaAddr(ch, partAddr);
+                        currStride = intraBCMixedMode.fencYuv->m_csize;
+                        ref = tmpPredYuv.getChromaAddr(ch, partAddr);
+                        refStride = tmpPredYuv.m_csize;
+                    }
+                    distAMVPBest += getSAD(ref, refStride, currStart, currStride, tempWidth, tempHeight);
+                }
+
+                MV check;
+                for (mvpIdxTemp = 0; mvpIdxTemp < AMVP_NUM_CANDS; mvpIdxTemp++)
+                {
+                    if (!(mvPred[mvpIdxTemp].x == check.x >> 2))
+                    {
+                        m_me.setMVP(bvPred[mvpIdxTemp]);
+                        bitsAMVPTemp = m_me.bitcost(cMv >> 2, bvPred[mvpIdxTemp]);
+                        if (bitsAMVPTemp < bitsAMVPBest)
+                        {
+                            bitsAMVPBest = bitsAMVPTemp;
+                            mvpIdxBest = mvpIdxTemp;
+                        }
+                    }
+                }
+
+                bitsAMVPBest++; // for MVP Index bits
+                costAMVPBest = distAMVPBest + m_rdCost.getCost(bitsAMVPBest);
+
+                MVField cMvFieldNeighboursIBC[MRG_MAX_NUM_CANDS][2]; // double length for mv of both lists
+                uint8_t uhInterDirNeighboursIBC[MRG_MAX_NUM_CANDS];
+                int numValidMergeCandIBC = 0;
+
+                if (ePartSize != SIZE_2Nx2N)
+                {
+                    if (0 && ePartSize != SIZE_2Nx2N && cu.m_cuDepth[0] >= 3)
+                    {
+                        cu.setPartSizeSubParts(SIZE_2Nx2N);
+                        if (partIdx == 0)
+                        {
+                            numValidMergeCandIBC = cu.getInterMergeCandidates(0, 0, cMvFieldNeighboursIBC, uhInterDirNeighboursIBC);
+                        }
+                        cu.setPartSizeSubParts(ePartSize);
+                    }
+                    else
+                    {
+                        numValidMergeCandIBC = cu.getInterMergeCandidates(pu.puAbsPartIdx, partIdx, cMvFieldNeighboursIBC, uhInterDirNeighboursIBC);
+                    }
+
+                    cu.roundMergeCandidates(cMvFieldNeighboursIBC, numValidMergeCandIBC);
+                    restrictBipredMergeCand(&cu, partIdx, cMvFieldNeighboursIBC, uhInterDirNeighboursIBC, numValidMergeCandIBC);
+
+                    for (mrgIdxTemp = 0; mrgIdxTemp < numValidMergeCandIBC; mrgIdxTemp++)
+                    {
+                        if (uhInterDirNeighboursIBC[mrgIdxTemp] != 1)
+                        {
+                            continue;
+                        }
+                        if (m_slice->m_refPOCList[0][cMvFieldNeighboursIBC[mrgIdxTemp][0].refIdx] != m_slice->m_poc)
+                        {
+                            continue;
+                        }
+
+                        if (!isBlockVectorValid(xCUStart + xStartInCU, yCUStart + yStartInCU, dummyWidth, dummyHeight, &cu,
+                            xStartInCU, yStartInCU, (cMvFieldNeighboursIBC[mrgIdxTemp][0].mv.x >> 2), (cMvFieldNeighboursIBC[mrgIdxTemp][0].mv.y >> 2), m_param->maxCUSize))
+                        {
+                            continue;
+                        }
+                        bitsMergeTemp = mrgIdxTemp == m_param->maxNumMergeCand ? mrgIdxTemp : mrgIdxTemp + 1;
+                        distMergeTemp = 0;
+
+                        cu.setPUMv(0, cMvFieldNeighboursIBC[mrgIdxTemp][0].mv, pu.puAbsPartIdx, partIdx);
+                        cu.setPURefIdx(0, (int8_t)(m_slice->m_numRefIdx[0] - 1), pu.puAbsPartIdx, partIdx);
+                        cu.setPUMv(1, MV(0, 0), pu.puAbsPartIdx, partIdx);
+                        cu.setPURefIdx(1, -1, pu.puAbsPartIdx, partIdx);
+                        cu.setPUInterDir(1, pu.puAbsPartIdx, partIdx);
+                        motionCompensation(cu, pu, tmpPredYuv, 1, 1);
+
+                        for (int ch = TEXT_LUMA; ch < MAX_NUM_COMPONENT; ch++)
+                        {
+                            int tempHeight, tempWidth;
+                            if (ch == 0)
+                            {
+                                tempHeight = dummyHeight;
+                                tempWidth = dummyWidth;
+                                currStart = intraBCMixedMode.fencYuv->getLumaAddr(partAddr);
+                                currStride = intraBCMixedMode.fencYuv->m_size;
+                                ref = tmpPredYuv.getLumaAddr(partAddr);
+                                refStride = tmpPredYuv.m_size;
+                            }
+                            else
+                            {
+                                tempHeight = dummyHeight >> m_vChromaShift;
+                                tempWidth = dummyWidth >> m_hChromaShift;
+
+                                currStart = intraBCMixedMode.fencYuv->getChromaAddr(ch, partAddr);
+                                currStride = intraBCMixedMode.fencYuv->m_csize;
+                                ref = tmpPredYuv.getChromaAddr(ch, partAddr);
+                                refStride = tmpPredYuv.m_csize;
+                            }
+                            distMergeTemp += getSAD(ref, refStride, currStart, currStride, tempWidth, tempHeight);
+                        }
+                        costMergeTemp = distMergeTemp + m_rdCost.getCost(bitsMergeTemp);
+
+                        if (costMergeTemp < costMergeBest)
+                        {
+                            costMergeBest = costMergeTemp;
+                            mrgIdxBest = mrgIdxTemp;
+                        }
+                    }
+                }
+
+                if (costMergeBest < costAMVPBest)
+                {
+                    cost[combo] += costMergeBest;
+                    isIBCMergeMode[combo] = true;
+                    bestIBCMvpIdx[combo] = mrgIdxBest;
+
+                    MVField mvField[2];
+                    MV mv(cMvFieldNeighboursIBC[mrgIdxBest][0].mv.x, cMvFieldNeighboursIBC[mrgIdxBest][0].mv.y);
+                    mvField[0].mv = mv;
+                    mvField[0].refIdx = m_slice->m_numRefIdx[0] - 1;   // the current picture is at the last position of list0
+                    mvField[1].mv = cMvZero;
+                    mvField[1].refIdx = -1;
+                    cMRGMvFieldIBC[combo][0] = mvField[0];
+                    cMRGMvFieldIBC[combo][1] = mvField[1];
+                }
+                else
+                {
+                    cost[combo] += costAMVPBest;
+                    isIBCMergeMode[combo] = false;
+                    bestIBCMvpIdx[combo] = mvpIdxBest;
+                    cMvPredCand[combo][partIdx].set(bvPred[mvpIdxBest].x << 2, bvPred[mvpIdxBest].y << 2);
+                }
+
+                cu.setPUInterDir(1, pu.puAbsPartIdx, partIdx);  // list 0 prediction
+                if (isIBCMergeMode[combo])
+                {
+                    cu.setPUMv(0, cMRGMvFieldIBC[combo][0].mv, pu.puAbsPartIdx, partIdx);
+                }
+                else
+                {
+                    cu.setPUMv(0, iMvCandList[8 + partIdx], pu.puAbsPartIdx, partIdx);
+                    cu.setPURefIdx(0, (int8_t)(m_slice->m_numRefIdx[0] - 1), pu.puAbsPartIdx, partIdx);
+                    cu.setPURefIdx(1, -1, pu.puAbsPartIdx, partIdx);
+                }
+                // ibc merge
+            }
+            else // is inter PU
+            {
+                uint32_t  costInterTemp = 0;
+                uint32_t  costInterBest = UINT32_MAX;
+                const pixel* currStart;
+                int currStride;
+                pixel* pCurr;
+                pixel* ref;
+                int refStride;
+                MergeData merge;
+                memset(&merge, 0, sizeof(merge));
+
+                for (int refList = 0; refList < numPredDir; refList++)
+                {
+                    uint32_t numRef = refList ? ((m_slice->m_numRefIdx[1] > 1) ? 2 : 1) : ((m_slice->m_numRefIdx[0] - 1 > 1) ? 2 : 1);
+                    for (int refIdx = 0; refIdx < numRef; refIdx++)
+                    {
+                        MV cMv = iMvCandList[4 * refList + 2 * refIdx + partIdx];
+
+                        cu.getNeighbourMV(partIdx, pu.puAbsPartIdx, intraBCMixedMode.interNeighbours);
+                        int numMvc = cu.getPMV(intraBCMixedMode.interNeighbours, refList, refIdx, intraBCMixedMode.amvpCand[refList][refIdx], mvc);
+                        int mvpIdx;
+
+                        uint32_t  tempCost0 = 0;
+                        uint32_t  tempCost1 = 0;
+                        mvPred[0] = intraBCMixedMode.amvpCand[refList][refIdx][0];
+                        mvPred[1] = intraBCMixedMode.amvpCand[refList][refIdx][1];
+
+                        m_me.setMVP(mvPred[0]);
+                        tempCost0 = m_me.bitcost(cMv, mvPred[0]);
+                        m_me.setMVP(mvPred[1]);
+                        tempCost1 = m_me.bitcost(cMv, mvPred[1]);
+                        if (tempCost1 < tempCost0)
+                        {
+                            mvpIdx = 1;
+                        }
+                        else
+                        {
+                            mvpIdx = 0;
+                        }
+                        uint32_t bitsTemp = m_listSelBits[refList] + MVP_IDX_BITS;
+                        bitsTemp += getTUBits(refIdx, numRef);
+
+                        m_me.setMVP(mvPred[mvpIdx]);
+                        if (0) //UseIntegerMv
+                        {
+                            cu.setPUMv(refList, (cMv >> 2) << 2, pu.puAbsPartIdx, partIdx);
+                        }
+                        else
+                        {
+                            cu.setPUMv(refList, cMv, pu.puAbsPartIdx, partIdx);
+                        }
+                        cu.setPURefIdx(refList, refIdx, pu.puAbsPartIdx, partIdx);
+                        cu.setPUInterDir(1 + refList, pu.puAbsPartIdx, partIdx);
+                        motionCompensation(cu, pu, tmpPredYuv, 1, 1);
+
+                        costInterTemp = 0;
+                        for (int ch = TEXT_LUMA; ch < MAX_NUM_COMPONENT; ch++)
+                        {
+                            int tempHeight, tempWidth;
+                            if (ch == 0)
+                            {
+                                tempHeight = dummyHeight;
+                                tempWidth = dummyWidth;
+                                currStart = intraBCMixedMode.fencYuv->getLumaAddr(partAddr);
+                                currStride = intraBCMixedMode.fencYuv->m_size;
+                                ref = tmpPredYuv.getLumaAddr(partAddr);
+                                refStride = tmpPredYuv.m_size;
+                            }
+                            else
+                            {
+                                tempHeight = dummyHeight >> m_vChromaShift;
+                                tempWidth = dummyWidth >> m_hChromaShift;
+
+                                currStart = intraBCMixedMode.fencYuv->getChromaAddr(ch, partAddr);
+                                currStride = intraBCMixedMode.fencYuv->m_csize;
+                                ref = tmpPredYuv.getChromaAddr(ch, partAddr);
+                                refStride = tmpPredYuv.m_csize;
+                            }
+                            costInterTemp += getSAD(ref, refStride, currStart, currStride, tempWidth, tempHeight);
+
+                            if (costInterTemp >= costInterBest)
+                            {
+                                break;
+                            }
+                        }
+                        cu.setPURefIdx(refList, -1, pu.puAbsPartIdx, partIdx);
+
+                        costInterTemp += m_me.bitcost(cMv, mvPred[mvpIdx]);
+                        costInterTemp += m_rdCost.getCost(bitsTemp);
+
+                        if (costInterTemp < costInterBest)
+                        {
+                            costInterBest = costInterTemp;
+                            bestInterMvpIdx[combo] = mvpIdx;
+                            bestInterDir[combo] = refList;
+                            bestRefIdx[combo] = refIdx;
+                            cMvPredCand[combo][partIdx] = mvPred[mvpIdx];
+                        }
+                    }
+                }
+
+                uint32_t MRGInterDir = 0;
+                uint32_t MRGIndex = 0;
+
+                // find Merge result
+                uint32_t MRGCost = UINT32_MAX;
+                cu.m_mergeFlag[pu.puAbsPartIdx] = true;
+
+                mergeEstimation(cu, cuGeom, pu, partIdx, merge);
+                MRGInterDir = merge.dir;
+                cMRGMvField[combo][0] = merge.mvField[0];
+                cMRGMvField[combo][1] = merge.mvField[1];
+                MRGIndex = merge.index;
+                cu.setPURefIdx(0, -1, pu.puAbsPartIdx, partIdx);
+                cu.setPURefIdx(1, -1, pu.puAbsPartIdx, partIdx);
+
+                if (MRGCost < costInterBest)
+                {
+                    costInterBest = MRGCost;
+                    isMergeMode[combo] = true;
+                    bestInterMvpIdx[combo] = MRGIndex;
+                    bestInterDir[combo] = MRGInterDir;
+                }
+
+                cost[combo] += costInterBest;
+                if (isMergeMode[combo])
+                {
+                    cu.setPUInterDir(bestInterDir[combo], pu.puAbsPartIdx, partIdx);
+                    cu.setPUMv(0, cMRGMvField[combo][0].mv, pu.puAbsPartIdx, partIdx);
+                    cu.setPURefIdx(0, cMRGMvField[combo][0].refIdx, pu.puAbsPartIdx, partIdx);
+                    cu.setPUMv(1, cMRGMvField[combo][1].mv, pu.puAbsPartIdx, partIdx);
+                    cu.setPURefIdx(1, cMRGMvField[combo][1].refIdx, pu.puAbsPartIdx, partIdx);
+                }
+                else
+                {
+                    int refListOpt = bestInterDir[combo];
+                    int refIdxOpt = bestRefIdx[combo];
+                    if (0) //UseIntegerMv
+                    {
+                        cu.setPUMv(refListOpt, (iMvCandList[partIdx + 2 * refIdxOpt + 4 * refListOpt] >> 2) << 2, pu.puAbsPartIdx, partIdx);
+                    }
+                    else
+                    {
+                        cu.setPUMv(refListOpt, iMvCandList[partIdx + 2 * refIdxOpt + 4 * refListOpt], pu.puAbsPartIdx, partIdx);
+                    }
+                    cu.setPURefIdx(refListOpt, refIdxOpt, pu.puAbsPartIdx, partIdx);
+                    cu.setPURefIdx(1 - refListOpt, -1, pu.puAbsPartIdx, partIdx);
+                    cu.setPUInterDir(1 + refListOpt, pu.puAbsPartIdx, partIdx);
+                    cu.m_mvpIdx[refListOpt][pu.puAbsPartIdx] = bestInterMvpIdx[combo];
+                }
+            }
+        } // for ipartIdx
+    } // for combo
+
+    if (IBCValidFlag > 1)
+    {
+        return false;
+    }
+
+    MV cMvd;
+    MV cMVFinal;
+    if (cost[0] <= cost[1])
+    {
+        int iDummyWidth1, iDummyHeight1;
+        uint32_t partAddr = 0;
+        uint32_t partIdx = 0;
+        cu.getPartIndexAndSize(partIdx, partAddr, iDummyWidth1, iDummyHeight1);
+
+        if (isIBCMergeMode[0])
+        {
+            cu.m_mergeFlag[partAddr] = true;
+            cu.m_mvpIdx[0][partAddr] = bestIBCMvpIdx[0];
+            cu.setPUInterDir(1, partAddr, partIdx);  // list 0 prediction
+            cu.setPUMv(0, cMRGMvFieldIBC[0][0].mv, partAddr, partIdx);
+            cu.setPURefIdx(0, cMRGMvFieldIBC[0][0].refIdx, partAddr, partIdx);
+            cu.setPUMv(1, cMRGMvFieldIBC[0][1].mv, partAddr, partIdx);
+            cu.setPURefIdx(1, cMRGMvFieldIBC[0][1].refIdx, partAddr, partIdx);
+
+            cu.m_mvd[0][partAddr] = cMvZero;
+            cu.m_mvd[1][partAddr] = cMvZero;
+        }
+        else
+        {
+            cu.m_mergeFlag[partAddr] = false;
+
+            cMvd.set((iMvCandList[8].x - cMvPredCand[0][0].x) >> 2, (iMvCandList[8].y - cMvPredCand[0][0].y) >> 2);
+            cu.setPUMv(0, iMvCandList[8], partAddr, partIdx);
+            cu.m_mvd[0][partAddr] = cMvd;
+            cu.m_mvpIdx[0][partAddr] = bestIBCMvpIdx[0];
+            cu.setPURefIdx(0, m_slice->m_numRefIdx[0] - 1, partAddr, partIdx);
+            cu.setPURefIdx(1, -1, partAddr, partIdx);
+            cu.setPUInterDir(1, partAddr, partIdx);  // list 0 prediction
+        }
+
+        partIdx = 1;
+        cu.getPartIndexAndSize(partIdx, partAddr, iDummyWidth1, iDummyHeight1);
+
+        if (isMergeMode[0])
+        {
+            cu.m_mergeFlag[partAddr] = true;
+            cu.m_mvpIdx[0][partAddr] = bestInterMvpIdx[0];
+            cu.setPUInterDir(bestInterDir[0], partAddr, partIdx);  // list 0 prediction
+            cu.setPUMv(0, cMRGMvField[0][0].mv, partAddr, partIdx);
+            cu.setPURefIdx(0, cMRGMvField[0][0].refIdx, partAddr, partIdx);
+            cu.setPUMv(1, cMRGMvField[0][1].mv, partAddr, partIdx);
+            cu.setPURefIdx(1, cMRGMvField[0][1].refIdx, partAddr, partIdx);
+
+            cu.m_mvd[0][partAddr] = cMvZero;
+
+            cu.m_mvd[1][partAddr] = cMvZero;
+        }
+        else
+        {
+            int refListOpt = bestInterDir[0];
+            int refIdxOpt = bestRefIdx[0];
+            if (0) //UseIntegerMv
+            {
+                cMvd.set(((iMvCandList[1 + 2 * refIdxOpt + 4 * refListOpt].x >> 2) - (cMvPredCand[0][1].x >> 2)), ((iMvCandList[1 + 2 * refIdxOpt + 4 * refListOpt].y >> 2) - (cMvPredCand[0][1].y >> 2)));
+                cu.setPUMv(refListOpt, (iMvCandList[1 + 2 * refIdxOpt + 4 * refListOpt] >> 2) << 2, partAddr, partIdx);
+            }
+            else
+            {
+                cMvd.set(iMvCandList[1 + 2 * refIdxOpt + 4 * refListOpt].x - cMvPredCand[0][1].x, iMvCandList[1 + 2 * refIdxOpt + 4 * refListOpt].y - cMvPredCand[0][1].y);
+                cu.setPUMv(refListOpt, iMvCandList[1 + 2 * refIdxOpt + 4 * refListOpt], partAddr, partIdx);
+            }
+            cu.m_mvd[refListOpt][partAddr] = cMvd;
+            cu.setPURefIdx(refListOpt, refIdxOpt, partAddr, partIdx);
+            cu.setPURefIdx(1 - refListOpt, -1, partAddr, partIdx);
+            cu.setPUInterDir(1 + refListOpt, partAddr, partIdx);
+            cu.m_mergeFlag[partAddr] = false;
+            cu.m_mvpIdx[refListOpt][partAddr] = bestInterMvpIdx[0];
+        }
+    }
+    else
+    {
+        int dummyWidth2, dummyHeight2;
+        uint32_t partAddr = 0;
+        uint32_t partIdx = 0;
+
+        cu.getPartIndexAndSize(partIdx, partAddr, dummyWidth2, dummyHeight2);
+
+        if (isMergeMode[1])
+        {
+            cu.m_mergeFlag[partAddr] = true;
+            cu.m_mvpIdx[0][partAddr] = bestInterMvpIdx[1];
+            cu.setPUInterDir(bestInterDir[1], partAddr, partIdx);  // list 0 prediction
+            cu.setPUMv(0, cMRGMvField[1][0].mv, partAddr, partIdx);
+            cu.setPURefIdx(0, cMRGMvField[1][0].refIdx, partAddr, partIdx);
+            cu.setPUMv(1, cMRGMvField[1][1].mv, partAddr, partIdx);
+            cu.setPURefIdx(1, cMRGMvField[1][1].refIdx, partAddr, partIdx);
+
+            cu.m_mvd[0][partAddr] = cMvZero;
+            cu.m_mvd[1][partAddr] = cMvZero;
+        }
+        else
+        {
+            int refListOpt = bestInterDir[1];
+            int refIdxOpt = bestRefIdx[1];
+            if (0)//UseIntegerMv
+            {
+                cMvd.set((iMvCandList[2 * refIdxOpt + 4 * refListOpt].x >> 2) - (cMvPredCand[1][0].x >> 2), (iMvCandList[2 * refIdxOpt + 4 * refListOpt].y >> 2) - (cMvPredCand[1][0].y >> 2));
+                cu.setPUMv(refListOpt, (iMvCandList[2 * refIdxOpt + 4 * refListOpt] >> 2) << 2, partAddr, partIdx);
+            }
+            else
+            {
+                cMvd.set(iMvCandList[2 * refIdxOpt + 4 * refListOpt].x - cMvPredCand[1][0].x, iMvCandList[2 * refIdxOpt + 4 * refListOpt].y - cMvPredCand[1][0].y);
+                cu.setPUMv(refListOpt, iMvCandList[2 * refIdxOpt + 4 * refListOpt], partAddr, partIdx);
+            }
+            cu.m_mvd[refListOpt][partAddr] = cMvd;
+            cu.setPURefIdx(refListOpt, refIdxOpt, partAddr, partIdx);
+            cu.setPURefIdx(1 - refListOpt, -1, partAddr, partIdx);
+            cu.setPUInterDir(1 + refListOpt, partAddr, partIdx);
+            cu.m_mergeFlag[partAddr] = false;
+            cu.m_mvpIdx[refListOpt][partAddr] = bestInterMvpIdx[1];
+        }
+
+        partIdx = 1;
+        cu.getPartIndexAndSize(partIdx, partAddr, dummyWidth2, dummyHeight2);
+
+        if (isIBCMergeMode[1])
+        {
+            cu.m_mergeFlag[partAddr] = true;
+            cu.m_mvpIdx[0][partAddr] = bestIBCMvpIdx[1];
+            cu.setPUInterDir(1, partAddr, partIdx);  // list 0 prediction
+            cu.setPUMv(0, cMRGMvFieldIBC[1][0].mv, partAddr, partIdx);
+            cu.setPURefIdx(0, cMRGMvFieldIBC[1][0].refIdx, partAddr, partIdx);
+            cu.setPUMv(1, cMRGMvFieldIBC[1][1].mv, partAddr, partIdx);
+            cu.setPURefIdx(1, cMRGMvFieldIBC[1][1].refIdx, partAddr, partIdx);
+
+            cu.m_mvd[0][partAddr] = cMvZero;
+            cu.m_mvd[1][partAddr] = cMvZero;
+        }
+        else
+        {
+            cu.m_mergeFlag[partAddr] = false;
+
+            cMvd.set(((iMvCandList[9].x - cMvPredCand[1][1].x) >> 2), (iMvCandList[9].y - cMvPredCand[1][1].y) >> 2);
+            cu.setPUMv(0, iMvCandList[9], partAddr, partIdx);
+            cu.m_mvd[0][partAddr] = cMvd;
+            cu.m_mvpIdx[0][partAddr] = bestIBCMvpIdx[1];
+            cu.setPURefIdx(0, m_slice->m_numRefIdx[0] - 1, partAddr, partIdx);
+            cu.setPURefIdx(1, -1, partAddr, partIdx);
+            cu.setPUInterDir(1, partAddr, partIdx);  // list 0 prediction
+        }
+    }
+    for (int partIdx = 0; partIdx < numPart; ++partIdx)
+    {
+        PredictionUnit pu(cu, cuGeom, partIdx);
+        motionCompensation(cu, pu, *predYuv, 1, 1);
+    }
+
+    return true;
+}
+
 void Search::getBlkBits(PartSize cuMode, bool bPSlice, int partIdx, uint32_t lastMode, uint32_t blockBit[3])
 {
     if (cuMode == SIZE_2Nx2N)
diff --git a/source/encoder/search.h b/source/encoder/search.h
index e90964619..e05d6ef95 100644
--- a/source/encoder/search.h
+++ b/source/encoder/search.h
@@ -290,6 +290,11 @@ public:
 
     int             m_ibcEnabled;
 
+    int             m_numBVs = 0;
+    int             m_numBV16s = 0;
+    MV              m_BVs[64];
+    uint32_t        m_lastCandCost;
+    MV              tempLastIntraBCMv[2] = { 0,0 };
 #if DETAILED_CU_STATS
     /* Accumulate CU statistics separately for each frame encoder */
     CUStats         m_stats[X265_MAX_FRAME_THREADS];
@@ -333,6 +338,23 @@ public:
 
     MV getLowresMV(const CUData& cu, const PredictionUnit& pu, int list, int ref);
 
+    bool      predIntraBCSearch(Mode& intraBCMode, const CUGeom& cuGeom, bool bChromaMC, PartSize ePartSize, bool testOnlyPred, bool bUse1DSearchFor8x8);
+    void      intraBlockCopyEstimate(Mode& intraBCMode, const CUGeom& cuGeom, int puIdx, MV* pred, MV& mv, uint32_t& cost, bool testOnlyPred, bool bUse1DSearchFor8x8);
+    void      setIntraSearchRange(Mode& intraBCMode, MV& pred, int puIdx, int roiWidth, int roiHeight, MV& searchRangeLT, MV& searchRangeRB);
+    void      intraPatternSearch(Mode& intraBCMode, const CUGeom& cuGeom, int puIdx, uint32_t partAddr, pixel* refY, int refStride, MV* searchRangeLT, MV* searchRangeRB,
+        MV& mv, uint32_t& cost, int roiwidth, int roiheight, bool testOnlyPred, bool bUse1DSearchFor8x8);
+    bool      isValidIntraBCSearchArea(CUData* cu, int predX, int predY, int roiWidth, int roiHeight, int partOffset);
+    bool      isBlockVectorValid(int xPos, int yPos, int width, int height, CUData* pcCU,
+        int xStartInCU, int yStartInCU, int xBv, int yBv, int ctuSize);
+    void      intraBCSearchMVCandUpdate(uint32_t sad, int x, int y, uint32_t* sadBestCand, MV* cMVCand);
+    void      updateBVMergeCandLists(int roiWidth, int roiHeight, MV* mvCand);
+    int       intraBCSearchMVChromaRefine(Mode& intraBCMode, const CUGeom& cuGeom, int roiWidth, int roiHeight, int cuPelX, int cuPelY, uint32_t* sadBestCand, MV* cMVCand,
+        uint32_t partOffset, int puIdx);
+    static    uint32_t mergeCandLists(MV* dst, uint32_t dn, MV* src, uint32_t sn, bool isSrcQuarPel);
+    uint32_t  getSAD(pixel* ref, int refStride, const pixel* curr, int currStride, int width, int height);
+    bool      predMixedIntraBCInterSearch(Mode& intraBCMode, const CUGeom& cuGeom, bool bChromaMC, PartSize ePartSize, bool testOnlyPred, MV* iMVCandList = (0, 0));
+    void      restrictBipredMergeCand(CUData* cu, uint32_t puIdx, MVField(*mvFieldNeighbours)[2], uint8_t* interDirNeighbours, int numValidMergeCand);
+
     class PME : public BondedTaskGroup
     {
     public:
diff --git a/source/x265.h b/source/x265.h
index abe3f16f0..889dda934 100644
--- a/source/x265.h
+++ b/source/x265.h
@@ -657,6 +657,10 @@ typedef enum
 /* SCC Extension Options */
 #define SCC_EXT_IDX               3
 #define NUM_EXTENSION_FLAGS       8
+#define SCM_S0067_NUM_CANDIDATES  64
+#define CHROMA_REFINEMENT_CANDIDATES  8
+#define SCM_S0067_IBC_FULL_1D_SEARCH_FOR_PU  2 ///< Do full horizontal/vertical search for Nx2N
+#define SCM_S0067_MAX_CAND_SIZE  32 ///< 32 or 64, 16 by default
 
 #define X265_IPRATIO_STRENGTH   1.43
 
-- 
2.36.0.windows.1

